# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from openfeature.schemas.protobuf.schema.v1 import schema_pb2 as openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2

GRPC_GENERATED_VERSION = '1.70.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in openfeature/schemas/protobuf/schema/v1/schema_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class ServiceStub(object):
    """Service defines the exposed rpcs of flagd
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ResolveAll = channel.unary_unary(
                '/schema.v1.Service/ResolveAll',
                request_serializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveAllRequest.SerializeToString,
                response_deserializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveAllResponse.FromString,
                _registered_method=True)
        self.ResolveBoolean = channel.unary_unary(
                '/schema.v1.Service/ResolveBoolean',
                request_serializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveBooleanRequest.SerializeToString,
                response_deserializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveBooleanResponse.FromString,
                _registered_method=True)
        self.ResolveString = channel.unary_unary(
                '/schema.v1.Service/ResolveString',
                request_serializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveStringRequest.SerializeToString,
                response_deserializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveStringResponse.FromString,
                _registered_method=True)
        self.ResolveFloat = channel.unary_unary(
                '/schema.v1.Service/ResolveFloat',
                request_serializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveFloatRequest.SerializeToString,
                response_deserializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveFloatResponse.FromString,
                _registered_method=True)
        self.ResolveInt = channel.unary_unary(
                '/schema.v1.Service/ResolveInt',
                request_serializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveIntRequest.SerializeToString,
                response_deserializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveIntResponse.FromString,
                _registered_method=True)
        self.ResolveObject = channel.unary_unary(
                '/schema.v1.Service/ResolveObject',
                request_serializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveObjectRequest.SerializeToString,
                response_deserializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveObjectResponse.FromString,
                _registered_method=True)
        self.EventStream = channel.unary_stream(
                '/schema.v1.Service/EventStream',
                request_serializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.EventStreamRequest.SerializeToString,
                response_deserializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.EventStreamResponse.FromString,
                _registered_method=True)


class ServiceServicer(object):
    """Service defines the exposed rpcs of flagd
    """

    def ResolveAll(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ResolveBoolean(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ResolveString(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ResolveFloat(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ResolveInt(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ResolveObject(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def EventStream(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'ResolveAll': grpc.unary_unary_rpc_method_handler(
                    servicer.ResolveAll,
                    request_deserializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveAllRequest.FromString,
                    response_serializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveAllResponse.SerializeToString,
            ),
            'ResolveBoolean': grpc.unary_unary_rpc_method_handler(
                    servicer.ResolveBoolean,
                    request_deserializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveBooleanRequest.FromString,
                    response_serializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveBooleanResponse.SerializeToString,
            ),
            'ResolveString': grpc.unary_unary_rpc_method_handler(
                    servicer.ResolveString,
                    request_deserializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveStringRequest.FromString,
                    response_serializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveStringResponse.SerializeToString,
            ),
            'ResolveFloat': grpc.unary_unary_rpc_method_handler(
                    servicer.ResolveFloat,
                    request_deserializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveFloatRequest.FromString,
                    response_serializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveFloatResponse.SerializeToString,
            ),
            'ResolveInt': grpc.unary_unary_rpc_method_handler(
                    servicer.ResolveInt,
                    request_deserializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveIntRequest.FromString,
                    response_serializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveIntResponse.SerializeToString,
            ),
            'ResolveObject': grpc.unary_unary_rpc_method_handler(
                    servicer.ResolveObject,
                    request_deserializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveObjectRequest.FromString,
                    response_serializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveObjectResponse.SerializeToString,
            ),
            'EventStream': grpc.unary_stream_rpc_method_handler(
                    servicer.EventStream,
                    request_deserializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.EventStreamRequest.FromString,
                    response_serializer=openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.EventStreamResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'schema.v1.Service', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('schema.v1.Service', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Service(object):
    """Service defines the exposed rpcs of flagd
    """

    @staticmethod
    def ResolveAll(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/schema.v1.Service/ResolveAll',
            openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveAllRequest.SerializeToString,
            openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveAllResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ResolveBoolean(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/schema.v1.Service/ResolveBoolean',
            openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveBooleanRequest.SerializeToString,
            openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveBooleanResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ResolveString(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/schema.v1.Service/ResolveString',
            openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveStringRequest.SerializeToString,
            openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveStringResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ResolveFloat(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/schema.v1.Service/ResolveFloat',
            openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveFloatRequest.SerializeToString,
            openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveFloatResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ResolveInt(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/schema.v1.Service/ResolveInt',
            openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveIntRequest.SerializeToString,
            openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveIntResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ResolveObject(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/schema.v1.Service/ResolveObject',
            openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveObjectRequest.SerializeToString,
            openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.ResolveObjectResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def EventStream(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/schema.v1.Service/EventStream',
            openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.EventStreamRequest.SerializeToString,
            openfeature_dot_schemas_dot_protobuf_dot_schema_dot_v1_dot_schema__pb2.EventStreamResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



import typing

from openfeature.evaluation_context import EvaluationContext
from openfeature.hook import Hook, HookContext, HookHints


class SyncMetadataHook(Hook):
    def __init__(self, context_supplier: typing.Callable[[], EvaluationContext]):
        self.context_supplier = context_supplier

    def before(
        self, hook_context: HookContext, hints: HookHints
    ) -> typing.Optional[EvaluationContext]:
        return self.context_supplier()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: openfeature/schemas/protobuf/flagd/evaluation/v1/evaluation.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'openfeature/schemas/protobuf/flagd/evaluation/v1/evaluation.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nAopenfeature/schemas/protobuf/flagd/evaluation/v1/evaluation.proto\x12\x13\x66lagd.evaluation.v1\x1a\x1cgoogle/protobuf/struct.proto\"=\n\x11ResolveAllRequest\x12(\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\"\xa3\x01\n\x12ResolveAllResponse\x12\x41\n\x05\x66lags\x18\x01 \x03(\x0b\x32\x32.flagd.evaluation.v1.ResolveAllResponse.FlagsEntry\x1aJ\n\nFlagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12+\n\x05value\x18\x02 \x01(\x0b\x32\x1c.flagd.evaluation.v1.AnyFlag:\x02\x38\x01\"\xaa\x01\n\x07\x41nyFlag\x12\x0e\n\x06reason\x18\x01 \x01(\t\x12\x0f\n\x07variant\x18\x02 \x01(\t\x12\x14\n\nbool_value\x18\x03 \x01(\x08H\x00\x12\x16\n\x0cstring_value\x18\x04 \x01(\tH\x00\x12\x16\n\x0c\x64ouble_value\x18\x05 \x01(\x01H\x00\x12/\n\x0cobject_value\x18\x06 \x01(\x0b\x32\x17.google.protobuf.StructH\x00\x42\x07\n\x05value\"S\n\x15ResolveBooleanRequest\x12\x10\n\x08\x66lag_key\x18\x01 \x01(\t\x12(\n\x07\x63ontext\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\"s\n\x16ResolveBooleanResponse\x12\r\n\x05value\x18\x01 \x01(\x08\x12\x0e\n\x06reason\x18\x02 \x01(\t\x12\x0f\n\x07variant\x18\x03 \x01(\t\x12)\n\x08metadata\x18\x04 \x01(\x0b\x32\x17.google.protobuf.Struct\"R\n\x14ResolveStringRequest\x12\x10\n\x08\x66lag_key\x18\x01 \x01(\t\x12(\n\x07\x63ontext\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\"r\n\x15ResolveStringResponse\x12\r\n\x05value\x18\x01 \x01(\t\x12\x0e\n\x06reason\x18\x02 \x01(\t\x12\x0f\n\x07variant\x18\x03 \x01(\t\x12)\n\x08metadata\x18\x04 \x01(\x0b\x32\x17.google.protobuf.Struct\"Q\n\x13ResolveFloatRequest\x12\x10\n\x08\x66lag_key\x18\x01 \x01(\t\x12(\n\x07\x63ontext\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\"q\n\x14ResolveFloatResponse\x12\r\n\x05value\x18\x01 \x01(\x01\x12\x0e\n\x06reason\x18\x02 \x01(\t\x12\x0f\n\x07variant\x18\x03 \x01(\t\x12)\n\x08metadata\x18\x04 \x01(\x0b\x32\x17.google.protobuf.Struct\"O\n\x11ResolveIntRequest\x12\x10\n\x08\x66lag_key\x18\x01 \x01(\t\x12(\n\x07\x63ontext\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\"o\n\x12ResolveIntResponse\x12\r\n\x05value\x18\x01 \x01(\x03\x12\x0e\n\x06reason\x18\x02 \x01(\t\x12\x0f\n\x07variant\x18\x03 \x01(\t\x12)\n\x08metadata\x18\x04 \x01(\x0b\x32\x17.google.protobuf.Struct\"R\n\x14ResolveObjectRequest\x12\x10\n\x08\x66lag_key\x18\x01 \x01(\t\x12(\n\x07\x63ontext\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\"\x8b\x01\n\x15ResolveObjectResponse\x12&\n\x05value\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x0e\n\x06reason\x18\x02 \x01(\t\x12\x0f\n\x07variant\x18\x03 \x01(\t\x12)\n\x08metadata\x18\x04 \x01(\x0b\x32\x17.google.protobuf.Struct\"J\n\x13\x45ventStreamResponse\x12\x0c\n\x04type\x18\x01 \x01(\t\x12%\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\"\x14\n\x12\x45ventStreamRequest2\xd9\x05\n\x07Service\x12_\n\nResolveAll\x12&.flagd.evaluation.v1.ResolveAllRequest\x1a\'.flagd.evaluation.v1.ResolveAllResponse\"\x00\x12k\n\x0eResolveBoolean\x12*.flagd.evaluation.v1.ResolveBooleanRequest\x1a+.flagd.evaluation.v1.ResolveBooleanResponse\"\x00\x12h\n\rResolveString\x12).flagd.evaluation.v1.ResolveStringRequest\x1a*.flagd.evaluation.v1.ResolveStringResponse\"\x00\x12\x65\n\x0cResolveFloat\x12(.flagd.evaluation.v1.ResolveFloatRequest\x1a).flagd.evaluation.v1.ResolveFloatResponse\"\x00\x12_\n\nResolveInt\x12&.flagd.evaluation.v1.ResolveIntRequest\x1a\'.flagd.evaluation.v1.ResolveIntResponse\"\x00\x12h\n\rResolveObject\x12).flagd.evaluation.v1.ResolveObjectRequest\x1a*.flagd.evaluation.v1.ResolveObjectResponse\"\x00\x12\x64\n\x0b\x45ventStream\x12\'.flagd.evaluation.v1.EventStreamRequest\x1a(.flagd.evaluation.v1.EventStreamResponse\"\x00\x30\x01\x42\xc6\x01\n%dev.openfeature.flagd.grpc.evaluationZ\x13\x66lagd/evaluation/v1\xaa\x02!OpenFeature.Flagd.Grpc.Evaluation\xca\x02\x32OpenFeature\\Providers\\Flagd\\Schema\\Grpc\\Evaluation\xea\x02.OpenFeature::Flagd::Provider::Grpc::Evaluationb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'openfeature.schemas.protobuf.flagd.evaluation.v1.evaluation_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n%dev.openfeature.flagd.grpc.evaluationZ\023flagd/evaluation/v1\252\002!OpenFeature.Flagd.Grpc.Evaluation\312\0022OpenFeature\\Providers\\Flagd\\Schema\\Grpc\\Evaluation\352\002.OpenFeature::Flagd::Provider::Grpc::Evaluation'
  _globals['_RESOLVEALLRESPONSE_FLAGSENTRY']._loaded_options = None
  _globals['_RESOLVEALLRESPONSE_FLAGSENTRY']._serialized_options = b'8\001'
  _globals['_RESOLVEALLREQUEST']._serialized_start=120
  _globals['_RESOLVEALLREQUEST']._serialized_end=181
  _globals['_RESOLVEALLRESPONSE']._serialized_start=184
  _globals['_RESOLVEALLRESPONSE']._serialized_end=347
  _globals['_RESOLVEALLRESPONSE_FLAGSENTRY']._serialized_start=273
  _globals['_RESOLVEALLRESPONSE_FLAGSENTRY']._serialized_end=347
  _globals['_ANYFLAG']._serialized_start=350
  _globals['_ANYFLAG']._serialized_end=520
  _globals['_RESOLVEBOOLEANREQUEST']._serialized_start=522
  _globals['_RESOLVEBOOLEANREQUEST']._serialized_end=605
  _globals['_RESOLVEBOOLEANRESPONSE']._serialized_start=607
  _globals['_RESOLVEBOOLEANRESPONSE']._serialized_end=722
  _globals['_RESOLVESTRINGREQUEST']._serialized_start=724
  _globals['_RESOLVESTRINGREQUEST']._serialized_end=806
  _globals['_RESOLVESTRINGRESPONSE']._serialized_start=808
  _globals['_RESOLVESTRINGRESPONSE']._serialized_end=922
  _globals['_RESOLVEFLOATREQUEST']._serialized_start=924
  _globals['_RESOLVEFLOATREQUEST']._serialized_end=1005
  _globals['_RESOLVEFLOATRESPONSE']._serialized_start=1007
  _globals['_RESOLVEFLOATRESPONSE']._serialized_end=1120
  _globals['_RESOLVEINTREQUEST']._serialized_start=1122
  _globals['_RESOLVEINTREQUEST']._serialized_end=1201
  _globals['_RESOLVEINTRESPONSE']._serialized_start=1203
  _globals['_RESOLVEINTRESPONSE']._serialized_end=1314
  _globals['_RESOLVEOBJECTREQUEST']._serialized_start=1316
  _globals['_RESOLVEOBJECTREQUEST']._serialized_end=1398
  _globals['_RESOLVEOBJECTRESPONSE']._serialized_start=1401
  _globals['_RESOLVEOBJECTRESPONSE']._serialized_end=1540
  _globals['_EVENTSTREAMRESPONSE']._serialized_start=1542
  _globals['_EVENTSTREAMRESPONSE']._serialized_end=1616
  _globals['_EVENTSTREAMREQUEST']._serialized_start=1618
  _globals['_EVENTSTREAMREQUEST']._serialized_end=1638
  _globals['_SERVICE']._serialized_start=1641
  _globals['_SERVICE']._serialized_end=2370
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
*
Flag definition sync API

This proto defines a simple API to synchronize a feature flag definition.
It supports establishing a stream for getting notifications about changes in a flag definition.
"""

import builtins
import google.protobuf.descriptor
import google.protobuf.message
import google.protobuf.struct_pb2
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class SyncFlagsRequest(google.protobuf.message.Message):
    """SyncFlagsRequest is the request initiating the server-streaming rpc.
    Implementations of Flagd providers and Flagd itself send this request, acting as the client.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PROVIDER_ID_FIELD_NUMBER: builtins.int
    SELECTOR_FIELD_NUMBER: builtins.int
    provider_id: builtins.str
    """Optional: A unique identifier for flagd(grpc client) initiating the request. The server implementations may
    utilize this identifier to uniquely identify, validate(ex:- enforce authentication/authorization) and filter
    flag configurations that it can expose to this request. This field is intended to be optional. However server
    implementations may enforce it.
       ex:- provider_id: flagd-weatherapp-sidecar
    """
    selector: builtins.str
    """Optional: A selector for the flag configuration request. The server implementation may utilize this to select
    flag configurations from a collection, select the source of the flag or combine this to any desired underlying
    filtering mechanism.
       ex:- selector: 'source=database,app=weatherapp'
    """
    def __init__(
        self,
        *,
        provider_id: builtins.str = ...,
        selector: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["provider_id", b"provider_id", "selector", b"selector"]) -> None: ...

global___SyncFlagsRequest = SyncFlagsRequest

@typing.final
class SyncFlagsResponse(google.protobuf.message.Message):
    """SyncFlagsResponse is the server response containing feature flag configurations and the state"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FLAG_CONFIGURATION_FIELD_NUMBER: builtins.int
    flag_configuration: builtins.str
    """flagd feature flag configuration. Must be validated to schema - https://raw.githubusercontent.com/open-feature/schemas/main/json/flagd-definitions.json"""
    def __init__(
        self,
        *,
        flag_configuration: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["flag_configuration", b"flag_configuration"]) -> None: ...

global___SyncFlagsResponse = SyncFlagsResponse

@typing.final
class FetchAllFlagsRequest(google.protobuf.message.Message):
    """FetchAllFlagsRequest is the request to fetch all flags. Clients send this request as the client in order to resync their internal state"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PROVIDER_ID_FIELD_NUMBER: builtins.int
    SELECTOR_FIELD_NUMBER: builtins.int
    provider_id: builtins.str
    """Optional: A unique identifier for clients initiating the request. The server implementations may
    utilize this identifier to uniquely identify, validate(ex:- enforce authentication/authorization) and filter
    flag configurations that it can expose to this request. This field is intended to be optional. However server
    implementations may enforce it.
       ex:- provider_id: flagd-weatherapp-sidecar
    """
    selector: builtins.str
    """Optional: A selector for the flag configuration request. The server implementation may utilize this to select
    flag configurations from a collection, select the source of the flag or combine this to any desired underlying
    filtering mechanism.
       ex:- selector: 'source=database,app=weatherapp'
    """
    def __init__(
        self,
        *,
        provider_id: builtins.str = ...,
        selector: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["provider_id", b"provider_id", "selector", b"selector"]) -> None: ...

global___FetchAllFlagsRequest = FetchAllFlagsRequest

@typing.final
class FetchAllFlagsResponse(google.protobuf.message.Message):
    """FetchAllFlagsResponse is the server response containing feature flag configurations"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FLAG_CONFIGURATION_FIELD_NUMBER: builtins.int
    flag_configuration: builtins.str
    """flagd feature flag configuration. Must be validated to schema - https://raw.githubusercontent.com/open-feature/schemas/main/json/flagd-definitions.json"""
    def __init__(
        self,
        *,
        flag_configuration: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["flag_configuration", b"flag_configuration"]) -> None: ...

global___FetchAllFlagsResponse = FetchAllFlagsResponse

@typing.final
class GetMetadataRequest(google.protobuf.message.Message):
    """GetMetadataRequest is the request for retrieving metadata from the sync service"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___GetMetadataRequest = GetMetadataRequest

@typing.final
class GetMetadataResponse(google.protobuf.message.Message):
    """GetMetadataResponse contains metadata from the sync service"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    METADATA_FIELD_NUMBER: builtins.int
    @property
    def metadata(self) -> google.protobuf.struct_pb2.Struct: ...
    def __init__(
        self,
        *,
        metadata: google.protobuf.struct_pb2.Struct | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["metadata", b"metadata"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["metadata", b"metadata"]) -> None: ...

global___GetMetadataResponse = GetMetadataResponse



import os
from pathlib import Path

import pytest

from openfeature import api
from openfeature.contrib.provider.flagd import FlagdProvider


@pytest.fixture()
def flagd_provider_client():
    provider = FlagdProvider()
    api.set_provider(provider)
    yield api.get_client()
    provider.shutdown()


def setup_flag_file(base_dir: str, flag_file: str) -> str:
    contents = (Path(__file__).parent / "../test-harness/flags" / flag_file).read_text()
    dst_path = os.path.join(base_dir, flag_file)
    with open(dst_path, "w") as dst_file:
        dst_file.write(contents)
    return dst_path



import logging
import typing
from dataclasses import dataclass

import mmh3
import semver

JsonPrimitive = typing.Union[str, bool, float, int]
JsonLogicArg = typing.Union[JsonPrimitive, typing.Sequence[JsonPrimitive]]

logger = logging.getLogger("openfeature.contrib")


@dataclass
class Fraction:
    variant: str
    weight: int = 1


def fractional(data: dict, *args: JsonLogicArg) -> typing.Optional[str]:
    if not args:
        logger.error("No arguments provided to fractional operator.")
        return None

    bucket_by = None
    if isinstance(args[0], str):
        bucket_by = args[0]
        args = args[1:]
    else:
        seed = data.get("$flagd", {}).get("flagKey", "")
        targeting_key = data.get("targetingKey")
        if not targeting_key:
            logger.error("No targetingKey provided for fractional shorthand syntax.")
            return None
        bucket_by = seed + targeting_key

    if not bucket_by:
        logger.error("No hashKey value resolved")
        return None

    hash_ratio = abs(mmh3.hash(bucket_by)) / (2**31 - 1)
    bucket = hash_ratio * 100

    total_weight = 0
    fractions = []
    try:
        for arg in args:
            fraction = _parse_fraction(arg)
            if fraction:
                fractions.append(fraction)
                total_weight += fraction.weight

    except ValueError:
        logger.debug(f"Invalid {args} configuration")
        return None

    range_end: float = 0
    for fraction in fractions:
        range_end += fraction.weight * 100 / total_weight
        if bucket < range_end:
            return fraction.variant
    return None


def _parse_fraction(arg: JsonLogicArg) -> Fraction:
    if not isinstance(arg, (tuple, list)) or not arg or len(arg) > 2:
        raise ValueError(
            "Fractional variant weights must be (str, int) tuple or [str] list"
        )

    if not isinstance(arg[0], str):
        raise ValueError(
            "Fractional variant identifier (first element) isn't of type 'str'"
        )

    if len(arg) >= 2 and not isinstance(arg[1], int):
        raise ValueError(
            "Fractional variant weight value (second element) isn't of type 'int'"
        )

    fraction = Fraction(variant=arg[0])
    if len(arg) >= 2:
        fraction.weight = arg[1]

    return fraction


def starts_with(data: dict, *args: JsonLogicArg) -> typing.Optional[bool]:
    def f(s1: str, s2: str) -> bool:
        return s1.startswith(s2)

    return string_comp(f, data, *args)


def ends_with(data: dict, *args: JsonLogicArg) -> typing.Optional[bool]:
    def f(s1: str, s2: str) -> bool:
        return s1.endswith(s2)

    return string_comp(f, data, *args)


def string_comp(
    comparator: typing.Callable[[str, str], bool], data: dict, *args: JsonLogicArg
) -> typing.Optional[bool]:
    if not args:
        logger.error("No arguments provided to string_comp operator.")
        return None
    if len(args) != 2:
        logger.error("Exactly 2 args expected for string_comp operator.")
        return None
    arg1, arg2 = args
    if not isinstance(arg1, str):
        logger.debug(f"incorrect argument for first argument, expected string: {arg1}")
        return False
    if not isinstance(arg2, str):
        logger.debug(f"incorrect argument for second argument, expected string: {arg2}")
        return False

    return comparator(arg1, arg2)


def sem_ver(data: dict, *args: JsonLogicArg) -> typing.Optional[bool]:  # noqa: C901
    if not args:
        logger.error("No arguments provided to sem_ver operator.")
        return None
    if len(args) != 3:
        logger.error("Exactly 3 args expected for sem_ver operator.")
        return None

    arg1, op, arg2 = args

    try:
        v1 = parse_version(arg1)
        v2 = parse_version(arg2)
    except ValueError as e:
        logger.exception(e)
        return None

    if op == "=":
        return v1 == v2
    elif op == "!=":
        return v1 != v2
    elif op == "<":
        return v1 < v2
    elif op == "<=":
        return v1 <= v2
    elif op == ">":
        return v1 > v2
    elif op == ">=":
        return v1 >= v2
    elif op == "^":
        return v1.major == v2.major
    elif op == "~":
        return v1.major == v2.major and v1.minor == v2.minor
    else:
        logger.error(f"Op not supported by sem_ver: {op}")
        return None


def parse_version(arg: typing.Any) -> semver.Version:
    version = str(arg)
    if version.startswith(("v", "V")):
        version = version[1:]

    return semver.Version.parse(version)



import typing

from openfeature.contrib.provider.flagd.resolvers.process.connector.file_watcher import (
    FileWatcher,
)
from openfeature.evaluation_context import EvaluationContext
from openfeature.event import ProviderEventDetails
from openfeature.exception import FlagNotFoundError, ParseError
from openfeature.flag_evaluation import FlagResolutionDetails, Reason

from ..config import Config
from .process.connector import FlagStateConnector
from .process.connector.grpc_watcher import GrpcWatcher
from .process.flags import FlagStore
from .process.targeting import targeting

T = typing.TypeVar("T")


class InProcessResolver:
    def __init__(
        self,
        config: Config,
        emit_provider_ready: typing.Callable[[ProviderEventDetails, dict], None],
        emit_provider_error: typing.Callable[[ProviderEventDetails], None],
        emit_provider_stale: typing.Callable[[ProviderEventDetails], None],
        emit_provider_configuration_changed: typing.Callable[
            [ProviderEventDetails], None
        ],
    ):
        self.config = config
        self.flag_store = FlagStore(emit_provider_configuration_changed)
        self.connector: FlagStateConnector = (
            FileWatcher(
                self.config, self.flag_store, emit_provider_ready, emit_provider_error
            )
            if self.config.offline_flag_source_path
            else GrpcWatcher(
                self.config,
                self.flag_store,
                emit_provider_ready,
                emit_provider_error,
                emit_provider_stale,
            )
        )

    def initialize(self, evaluation_context: EvaluationContext) -> None:
        self.connector.initialize(evaluation_context)

    def shutdown(self) -> None:
        self.connector.shutdown()

    def resolve_boolean_details(
        self,
        key: str,
        default_value: bool,
        evaluation_context: typing.Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[bool]:
        return self._resolve(key, default_value, evaluation_context)

    def resolve_string_details(
        self,
        key: str,
        default_value: str,
        evaluation_context: typing.Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[str]:
        return self._resolve(key, default_value, evaluation_context)

    def resolve_float_details(
        self,
        key: str,
        default_value: float,
        evaluation_context: typing.Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[float]:
        result = self._resolve(key, default_value, evaluation_context)
        if isinstance(result.value, int):
            result.value = float(result.value)
        return result

    def resolve_integer_details(
        self,
        key: str,
        default_value: int,
        evaluation_context: typing.Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[int]:
        return self._resolve(key, default_value, evaluation_context)

    def resolve_object_details(
        self,
        key: str,
        default_value: typing.Union[dict, list],
        evaluation_context: typing.Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[typing.Union[dict, list]]:
        return self._resolve(key, default_value, evaluation_context)

    def _resolve(
        self,
        key: str,
        default_value: T,
        evaluation_context: typing.Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[T]:
        flag = self.flag_store.get_flag(key)
        if not flag:
            raise FlagNotFoundError(f"Flag with key {key} not present in flag store.")

        if flag.state == "DISABLED":
            return FlagResolutionDetails(default_value, reason=Reason.DISABLED)

        if not flag.targeting:
            variant, value = flag.default
            return FlagResolutionDetails(value, variant=variant, reason=Reason.STATIC)

        variant = targeting(flag.key, flag.targeting, evaluation_context)

        if variant is None:
            variant, value = flag.default
            return FlagResolutionDetails(value, variant=variant, reason=Reason.DEFAULT)
        if not isinstance(variant, (str, bool)):
            raise ParseError(
                "Parsed JSONLogic targeting did not return a string or bool"
            )

        variant, value = flag.get_variant(variant)
        if not value:
            raise ParseError(f"Resolved variant {variant} not in variants config.")

        return FlagResolutionDetails(
            value,
            variant=variant,
            reason=Reason.TARGETING_MATCH,
        )



import pytest

# not sure if we still need this test, as this is also covered with gherkin tests.
from openfeature.contrib.provider.flagd.config import (
    DEFAULT_CACHE,
    DEFAULT_CACHE_SIZE,
    DEFAULT_DEADLINE,
    DEFAULT_HOST,
    DEFAULT_KEEP_ALIVE,
    DEFAULT_OFFLINE_SOURCE_PATH,
    DEFAULT_PORT_IN_PROCESS,
    DEFAULT_PORT_RPC,
    DEFAULT_RESOLVER_TYPE,
    DEFAULT_RETRY_BACKOFF,
    DEFAULT_STREAM_DEADLINE,
    DEFAULT_TLS,
    ENV_VAR_CACHE_SIZE,
    ENV_VAR_CACHE_TYPE,
    ENV_VAR_DEADLINE_MS,
    ENV_VAR_HOST,
    ENV_VAR_KEEP_ALIVE_TIME_MS,
    ENV_VAR_OFFLINE_FLAG_SOURCE_PATH,
    ENV_VAR_PORT,
    ENV_VAR_RETRY_BACKOFF_MS,
    ENV_VAR_STREAM_DEADLINE_MS,
    ENV_VAR_TLS,
    CacheType,
    Config,
    ResolverType,
)


def test_return_default_values_rpc():
    config = Config()
    assert config.cache == DEFAULT_CACHE
    assert config.max_cache_size == DEFAULT_CACHE_SIZE
    assert config.deadline_ms == DEFAULT_DEADLINE
    assert config.host == DEFAULT_HOST
    assert config.keep_alive_time == DEFAULT_KEEP_ALIVE
    assert config.offline_flag_source_path == DEFAULT_OFFLINE_SOURCE_PATH
    assert config.port == DEFAULT_PORT_RPC
    assert config.resolver == DEFAULT_RESOLVER_TYPE
    assert config.retry_backoff_ms == DEFAULT_RETRY_BACKOFF
    assert config.stream_deadline_ms == DEFAULT_STREAM_DEADLINE
    assert config.tls is DEFAULT_TLS


def test_return_default_values_in_process():
    config = Config(resolver=ResolverType.IN_PROCESS)
    assert config.cache == DEFAULT_CACHE
    assert config.max_cache_size == DEFAULT_CACHE_SIZE
    assert config.deadline_ms == DEFAULT_DEADLINE
    assert config.host == DEFAULT_HOST
    assert config.keep_alive_time == DEFAULT_KEEP_ALIVE
    assert config.offline_flag_source_path == DEFAULT_OFFLINE_SOURCE_PATH
    assert config.port == DEFAULT_PORT_IN_PROCESS
    assert config.resolver == ResolverType.IN_PROCESS
    assert config.retry_backoff_ms == DEFAULT_RETRY_BACKOFF
    assert config.stream_deadline_ms == DEFAULT_STREAM_DEADLINE
    assert config.tls is DEFAULT_TLS


@pytest.fixture(params=ResolverType, scope="module")
def resolver_type(request):
    return request.param


def test_overrides_defaults_with_environment(monkeypatch, resolver_type):  # noqa: PLR0915
    cache = CacheType.DISABLED
    cache_size = 456
    deadline = 1
    host = "flagd"
    keep_alive = 2
    offline_path = "path"
    port = 1234
    retry_backoff = 3
    stream_deadline = 4
    tls = True

    monkeypatch.setenv(ENV_VAR_CACHE_TYPE, str(cache.value))
    monkeypatch.setenv(ENV_VAR_CACHE_SIZE, str(cache_size))
    monkeypatch.setenv(ENV_VAR_DEADLINE_MS, str(deadline))
    monkeypatch.setenv(ENV_VAR_HOST, host)
    monkeypatch.setenv(ENV_VAR_KEEP_ALIVE_TIME_MS, str(keep_alive))
    monkeypatch.setenv(ENV_VAR_OFFLINE_FLAG_SOURCE_PATH, offline_path)
    monkeypatch.setenv(ENV_VAR_PORT, str(port))
    monkeypatch.setenv(ENV_VAR_RETRY_BACKOFF_MS, str(retry_backoff))
    monkeypatch.setenv(ENV_VAR_STREAM_DEADLINE_MS, str(stream_deadline))
    monkeypatch.setenv(ENV_VAR_TLS, str(tls))

    config = Config()
    assert config.cache == cache
    assert config.max_cache_size == cache_size
    assert config.deadline_ms == deadline
    assert config.host == host
    assert config.keep_alive_time == keep_alive
    assert config.offline_flag_source_path == offline_path
    assert config.port == port
    assert config.retry_backoff_ms == retry_backoff
    assert config.stream_deadline_ms == stream_deadline
    assert config.tls is tls


def test_uses_arguments_over_environments_and_defaults(monkeypatch, resolver_type):  # noqa: PLR0915
    cache = CacheType.LRU
    cache_size = 456
    deadline = 1
    host = "flagd"
    keep_alive = 2
    offline_path = "path"
    port = 1234
    retry_backoff = 3
    stream_deadline = 4
    tls = True

    monkeypatch.setenv(ENV_VAR_CACHE_TYPE, str(cache.value) + "value")
    monkeypatch.setenv(ENV_VAR_CACHE_SIZE, str(cache_size) + "value")
    monkeypatch.setenv(ENV_VAR_DEADLINE_MS, str(deadline) + "value")
    monkeypatch.setenv(ENV_VAR_HOST, host + "value")
    monkeypatch.setenv(ENV_VAR_KEEP_ALIVE_TIME_MS, str(keep_alive) + "value")
    monkeypatch.setenv(ENV_VAR_OFFLINE_FLAG_SOURCE_PATH, offline_path + "value")
    monkeypatch.setenv(ENV_VAR_PORT, str(port) + "value")
    monkeypatch.setenv(ENV_VAR_RETRY_BACKOFF_MS, str(retry_backoff) + "value")
    monkeypatch.setenv(ENV_VAR_STREAM_DEADLINE_MS, str(stream_deadline) + "value")
    monkeypatch.setenv(ENV_VAR_TLS, str(tls) + "value")

    config = Config(
        cache=cache,
        max_cache_size=cache_size,
        deadline_ms=deadline,
        host=host,
        port=port,
        resolver=resolver_type,
        retry_backoff_ms=retry_backoff,
        stream_deadline_ms=stream_deadline,
        tls=tls,
        keep_alive_time=keep_alive,
        offline_flag_source_path=offline_path,
    )
    assert config.cache == cache
    assert config.max_cache_size == cache_size
    assert config.deadline_ms == deadline
    assert config.host == host
    assert config.keep_alive_time == keep_alive
    assert config.offline_flag_source_path == offline_path
    assert config.port == port
    assert config.retry_backoff_ms == retry_backoff
    assert config.stream_deadline_ms == stream_deadline
    assert config.tls is tls



import os
import time

import pytest

from openfeature import api
from openfeature.contrib.provider.flagd import FlagdProvider
from openfeature.contrib.provider.flagd.config import ResolverType
from openfeature.evaluation_context import EvaluationContext
from openfeature.event import ProviderEvent
from openfeature.exception import ErrorCode
from openfeature.flag_evaluation import Reason


def create_client(provider: FlagdProvider):
    api.set_provider(provider)
    return api.get_client()


@pytest.mark.parametrize(
    "file_name",
    [
        "not-a-flag.json",
        "basic-flag-wrong-structure.json",
        "basic-flag-invalid.not-json",
        "basic-flag-wrong-variant.json",
        "basic-flag-broken-state.json",
        "basic-flag-broken-variants.json",
        "basic-flag-broken-default.json",
        "basic-flag-broken-targeting.json",
    ],
)
def test_file_load_errors(file_name: str):
    path = os.path.abspath(os.path.join(os.path.dirname(__file__), "./flags/"))
    client = create_client(
        FlagdProvider(
            resolver_type=ResolverType.IN_PROCESS,
            offline_flag_source_path=f"{path}/{file_name}",
        )
    )

    res = client.get_boolean_details("basic-flag", False)

    assert res.value is False
    assert res.reason == Reason.ERROR
    assert res.error_code == ErrorCode.FLAG_NOT_FOUND


@pytest.mark.parametrize(
    "file_name",
    [
        "invalid-semver-op.json",
        "invalid-semver-args.json",
        "invalid-stringcomp-args.json",
        "invalid-fractional-args.json",
        "invalid-fractional-args-wrong-content.json",
        "invalid-fractional-weights.json",
        "invalid-fractional-weights-strings.json",
    ],
)
def test_json_logic_parse_errors(file_name: str):
    path = os.path.abspath(os.path.join(os.path.dirname(__file__), "./flags/"))
    client = create_client(
        FlagdProvider(
            resolver_type=ResolverType.IN_PROCESS,
            offline_flag_source_path=f"{path}/{file_name}",
        )
    )

    res = client.get_string_details("basic-flag", "fallback", EvaluationContext("123"))

    assert res.value == "default"
    assert res.reason == Reason.DEFAULT


def test_flag_disabled():
    path = os.path.abspath(os.path.join(os.path.dirname(__file__), "./flags/"))
    client = create_client(
        FlagdProvider(
            resolver_type=ResolverType.IN_PROCESS,
            offline_flag_source_path=f"{path}/basic-flag-disabled.json",
        )
    )

    res = client.get_string_details("basic-flag", "fallback", EvaluationContext("123"))

    assert res.value == "fallback"
    assert res.reason == Reason.DISABLED


@pytest.mark.parametrize("wait", (500, 250))
def test_grpc_sync_fail_deadline(wait: int):
    init_failed = False

    def fail(*args, **kwargs):
        nonlocal init_failed
        init_failed = True

    api.get_client().add_handler(ProviderEvent.PROVIDER_ERROR, fail)

    t = time.time()
    api.set_provider(
        FlagdProvider(
            resolver_type=ResolverType.IN_PROCESS,
            port=99999,  # dead port to test failure
            deadline_ms=wait,
        )
    )

    elapsed = time.time() - t
    assert abs(elapsed - wait * 0.001) < 0.11
    assert init_failed



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
*
DEPRECATED; use flagd.sync.v1
Flag definition sync API

This proto defines a simple API to synchronize a feature flag definition.
It supports establishing a stream for getting notifications about changes in a flag definition.
"""

import builtins
import google.protobuf.descriptor
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _SyncState:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _SyncStateEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_SyncState.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    SYNC_STATE_UNSPECIFIED: _SyncState.ValueType  # 0
    """Value is ignored by the listening flagd"""
    SYNC_STATE_ALL: _SyncState.ValueType  # 1
    """All the flags matching the request. This is the default response and other states can be ignored
    by the implementation. Flagd internally replaces all existing flags for this response state.
    """
    SYNC_STATE_ADD: _SyncState.ValueType  # 2
    """Convey an addition of a flag. Flagd internally handles this by combining new flags with existing ones"""
    SYNC_STATE_UPDATE: _SyncState.ValueType  # 3
    """Convey an update of a flag. Flagd internally attempts to update if the updated flag already exist OR if it does not,
    it will get added
    """
    SYNC_STATE_DELETE: _SyncState.ValueType  # 4
    """Convey a deletion of a flag. Flagd internally removes the flag"""
    SYNC_STATE_PING: _SyncState.ValueType  # 5
    """Optional server ping to check client connectivity. Handling is ignored by flagd and is to merely support live check"""

class SyncState(_SyncState, metaclass=_SyncStateEnumTypeWrapper):
    """SyncState conveys the state of the payload. These states are related to flagd isync.go type definitions but
    contains extras to optimize grpc use case. Refer - https://github.com/open-feature/flagd/blob/main/pkg/sync/isync.go
    """

SYNC_STATE_UNSPECIFIED: SyncState.ValueType  # 0
"""Value is ignored by the listening flagd"""
SYNC_STATE_ALL: SyncState.ValueType  # 1
"""All the flags matching the request. This is the default response and other states can be ignored
by the implementation. Flagd internally replaces all existing flags for this response state.
"""
SYNC_STATE_ADD: SyncState.ValueType  # 2
"""Convey an addition of a flag. Flagd internally handles this by combining new flags with existing ones"""
SYNC_STATE_UPDATE: SyncState.ValueType  # 3
"""Convey an update of a flag. Flagd internally attempts to update if the updated flag already exist OR if it does not,
it will get added
"""
SYNC_STATE_DELETE: SyncState.ValueType  # 4
"""Convey a deletion of a flag. Flagd internally removes the flag"""
SYNC_STATE_PING: SyncState.ValueType  # 5
"""Optional server ping to check client connectivity. Handling is ignored by flagd and is to merely support live check"""
global___SyncState = SyncState

@typing.final
class SyncFlagsRequest(google.protobuf.message.Message):
    """SyncFlagsRequest is the request initiating the sever-streaming rpc. Flagd sends this request, acting as the client"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PROVIDER_ID_FIELD_NUMBER: builtins.int
    SELECTOR_FIELD_NUMBER: builtins.int
    provider_id: builtins.str
    """Optional: A unique identifier for flagd(grpc client) initiating the request. The server implementations may
    utilize this identifier to uniquely identify, validate(ex:- enforce authentication/authorization) and filter
    flag configurations that it can expose to this request. This field is intended to be optional. However server
    implementations may enforce it.
       ex:- provider_id: flagd-weatherapp-sidecar
    """
    selector: builtins.str
    """Optional: A selector for the flag configuration request. The server implementation may utilize this to select
    flag configurations from a collection, select the source of the flag or combine this to any desired underlying
    filtering mechanism.
       ex:- selector: 'source=database,app=weatherapp'
    """
    def __init__(
        self,
        *,
        provider_id: builtins.str = ...,
        selector: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["provider_id", b"provider_id", "selector", b"selector"]) -> None: ...

global___SyncFlagsRequest = SyncFlagsRequest

@typing.final
class SyncFlagsResponse(google.protobuf.message.Message):
    """SyncFlagsResponse is the server response containing feature flag configurations and the state"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FLAG_CONFIGURATION_FIELD_NUMBER: builtins.int
    STATE_FIELD_NUMBER: builtins.int
    flag_configuration: builtins.str
    """flagd feature flag configuration. Must be validated to schema - https://raw.githubusercontent.com/open-feature/schemas/main/json/flagd-definitions.json"""
    state: global___SyncState.ValueType
    """State conveying the operation to be performed by flagd. See the descriptions of SyncState for an explanation of
    supported values
    """
    def __init__(
        self,
        *,
        flag_configuration: builtins.str = ...,
        state: global___SyncState.ValueType = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["flag_configuration", b"flag_configuration", "state", b"state"]) -> None: ...

global___SyncFlagsResponse = SyncFlagsResponse

@typing.final
class FetchAllFlagsRequest(google.protobuf.message.Message):
    """FetchAllFlagsRequest is the request to fetch all flags. Flagd sends this request as the client in order to resync its internal state"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PROVIDER_ID_FIELD_NUMBER: builtins.int
    SELECTOR_FIELD_NUMBER: builtins.int
    provider_id: builtins.str
    """Optional: A unique identifier for flagd(grpc client) initiating the request. The server implementations may
    utilize this identifier to uniquely identify, validate(ex:- enforce authentication/authorization) and filter
    flag configurations that it can expose to this request. This field is intended to be optional. However server
    implementations may enforce it.
       ex:- provider_id: flagd-weatherapp-sidecar
    """
    selector: builtins.str
    """Optional: A selector for the flag configuration request. The server implementation may utilize this to select
    flag configurations from a collection, select the source of the flag or combine this to any desired underlying
    filtering mechanism.
       ex:- selector: 'source=database,app=weatherapp'
    """
    def __init__(
        self,
        *,
        provider_id: builtins.str = ...,
        selector: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["provider_id", b"provider_id", "selector", b"selector"]) -> None: ...

global___FetchAllFlagsRequest = FetchAllFlagsRequest

@typing.final
class FetchAllFlagsResponse(google.protobuf.message.Message):
    """ FetchAllFlagsResponse is the server response containing feature flag configurations"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FLAG_CONFIGURATION_FIELD_NUMBER: builtins.int
    flag_configuration: builtins.str
    """flagd feature flag configuration. Must be validated to schema - https://raw.githubusercontent.com/open-feature/schemas/main/json/flagd-definitions.json"""
    def __init__(
        self,
        *,
        flag_configuration: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["flag_configuration", b"flag_configuration"]) -> None: ...

global___FetchAllFlagsResponse = FetchAllFlagsResponse



import logging
import threading
import time
import typing

import grpc
from cachebox import BaseCacheImpl, LRUCache
from google.protobuf.json_format import MessageToDict
from google.protobuf.struct_pb2 import Struct
from grpc import ChannelConnectivity

from openfeature.evaluation_context import EvaluationContext
from openfeature.event import ProviderEventDetails
from openfeature.exception import (
    ErrorCode,
    FlagNotFoundError,
    GeneralError,
    InvalidContextError,
    ParseError,
    ProviderNotReadyError,
    TypeMismatchError,
)
from openfeature.flag_evaluation import FlagResolutionDetails, Reason
from openfeature.schemas.protobuf.flagd.evaluation.v1 import (
    evaluation_pb2,
    evaluation_pb2_grpc,
)

from ..config import CacheType, Config
from ..flag_type import FlagType

if typing.TYPE_CHECKING:
    from google.protobuf.message import Message

T = typing.TypeVar("T")

logger = logging.getLogger("openfeature.contrib")


class GrpcResolver:
    def __init__(
        self,
        config: Config,
        emit_provider_ready: typing.Callable[[ProviderEventDetails], None],
        emit_provider_error: typing.Callable[[ProviderEventDetails], None],
        emit_provider_stale: typing.Callable[[ProviderEventDetails], None],
        emit_provider_configuration_changed: typing.Callable[
            [ProviderEventDetails], None
        ],
    ):
        self.active = False
        self.config = config
        self.emit_provider_ready = emit_provider_ready
        self.emit_provider_error = emit_provider_error
        self.emit_provider_stale = emit_provider_stale
        self.emit_provider_configuration_changed = emit_provider_configuration_changed
        self.cache: typing.Optional[BaseCacheImpl] = (
            LRUCache(maxsize=self.config.max_cache_size)
            if self.config.cache == CacheType.LRU
            else None
        )

        self.retry_grace_period = config.retry_grace_period
        self.streamline_deadline_seconds = config.stream_deadline_ms * 0.001
        self.deadline = config.deadline_ms * 0.001
        self.connected = False
        self.channel = self._generate_channel(config)
        self.stub = evaluation_pb2_grpc.ServiceStub(self.channel)

        self.thread: typing.Optional[threading.Thread] = None
        self.timer: typing.Optional[threading.Timer] = None

        self.start_time = time.time()

    def _generate_channel(self, config: Config) -> grpc.Channel:
        target = f"{config.host}:{config.port}"
        # Create the channel with the service config
        options = [
            ("grpc.keepalive_time_ms", config.keep_alive_time),
            ("grpc.initial_reconnect_backoff_ms", config.retry_backoff_ms),
            ("grpc.max_reconnect_backoff_ms", config.retry_backoff_max_ms),
            ("grpc.min_reconnect_backoff_ms", config.deadline_ms),
        ]
        if config.tls:
            channel_args = {
                "options": options,
                "credentials": grpc.ssl_channel_credentials(),
            }
            if config.cert_path:
                with open(config.cert_path, "rb") as f:
                    channel_args["credentials"] = grpc.ssl_channel_credentials(f.read())

            channel = grpc.secure_channel(target, **channel_args)

        else:
            channel = grpc.insecure_channel(
                target,
                options=options,
            )

        return channel

    def initialize(self, evaluation_context: EvaluationContext) -> None:
        self.connect()

    def shutdown(self) -> None:
        self.active = False
        self.channel.unsubscribe(self._state_change_callback)
        self.channel.close()
        if self.timer and self.timer.is_alive():
            logger.debug("gRPC error timer cancelled due to shutdown")
            self.timer.cancel()
        if self.cache:
            self.cache.clear()

    def connect(self) -> None:
        self.active = True

        # Run monitoring in a separate thread
        self.monitor_thread = threading.Thread(
            target=self.monitor, daemon=True, name="FlagdGrpcServiceMonitorThread"
        )
        self.monitor_thread.start()
        ## block until ready or deadline reached
        timeout = self.deadline + time.time()
        while not self.connected and time.time() < timeout:
            time.sleep(0.05)
        logger.debug("Finished blocking gRPC state initialization")

        if not self.connected:
            raise ProviderNotReadyError(
                "Blocking init finished before data synced. Consider increasing startup deadline to avoid inconsistent evaluations."
            )

    def monitor(self) -> None:
        self.channel.subscribe(self._state_change_callback, try_to_connect=True)

    def _state_change_callback(self, new_state: ChannelConnectivity) -> None:
        logger.debug(f"gRPC state change: {new_state}")
        if (
            new_state == grpc.ChannelConnectivity.READY
            or new_state == grpc.ChannelConnectivity.IDLE
        ):
            if not self.thread or not self.thread.is_alive():
                self.thread = threading.Thread(
                    target=self.listen,
                    daemon=True,
                    name="FlagdGrpcServiceWorkerThread",
                )
                self.thread.start()

            if self.timer and self.timer.is_alive():
                logger.debug("gRPC error timer expired")
                self.timer.cancel()

        elif new_state == ChannelConnectivity.TRANSIENT_FAILURE:
            # this is the failed reconnect attempt so we are going into stale
            self.emit_provider_stale(
                ProviderEventDetails(
                    message="gRPC sync disconnected, reconnecting",
                )
            )
            self.start_time = time.time()
            # adding a timer, so we can emit the error event after time
            self.timer = threading.Timer(self.retry_grace_period, self.emit_error)

            logger.debug("gRPC error timer started")
            self.timer.start()
            self.connected = False

    def emit_error(self) -> None:
        logger.debug("gRPC error emitted")
        if self.cache:
            self.cache.clear()
        self.emit_provider_error(
            ProviderEventDetails(
                message="gRPC sync disconnected, reconnecting",
                error_code=ErrorCode.GENERAL,
            )
        )

    def listen(self) -> None:
        logger.debug("gRPC starting listener thread")
        call_args = (
            {"timeout": self.streamline_deadline_seconds}
            if self.streamline_deadline_seconds > 0
            else {}
        )
        request = evaluation_pb2.EventStreamRequest()

        # defining a never ending loop to recreate the stream
        while self.active:
            try:
                logger.debug("Setting up gRPC sync flags connection")
                for message in self.stub.EventStream(
                    request, wait_for_ready=True, **call_args
                ):
                    if message.type == "provider_ready":
                        self.emit_provider_ready(
                            ProviderEventDetails(
                                message="gRPC sync connection established"
                            )
                        )
                        self.connected = True
                    elif message.type == "configuration_change":
                        data = MessageToDict(message)["data"]
                        self.handle_changed_flags(data)

                    if not self.active:
                        logger.info("Terminating gRPC sync thread")
                        return
            except grpc.RpcError as e:  # noqa: PERF203
                # although it seems like this error log is not interesting, without it, the retry is not working as expected
                logger.debug(f"SyncFlags stream error, {e.code()=} {e.details()=}")
            except ParseError:
                logger.exception(
                    f"Could not parse flag data using flagd syntax: {message=}"
                )

    def handle_changed_flags(self, data: typing.Any) -> None:
        changed_flags = list(data["flags"].keys())

        if self.cache:
            for flag in changed_flags:
                self.cache.pop(flag)

        self.emit_provider_configuration_changed(ProviderEventDetails(changed_flags))

    def resolve_boolean_details(
        self,
        key: str,
        default_value: bool,
        evaluation_context: typing.Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[bool]:
        return self._resolve(key, FlagType.BOOLEAN, default_value, evaluation_context)

    def resolve_string_details(
        self,
        key: str,
        default_value: str,
        evaluation_context: typing.Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[str]:
        return self._resolve(key, FlagType.STRING, default_value, evaluation_context)

    def resolve_float_details(
        self,
        key: str,
        default_value: float,
        evaluation_context: typing.Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[float]:
        return self._resolve(key, FlagType.FLOAT, default_value, evaluation_context)

    def resolve_integer_details(
        self,
        key: str,
        default_value: int,
        evaluation_context: typing.Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[int]:
        return self._resolve(key, FlagType.INTEGER, default_value, evaluation_context)

    def resolve_object_details(
        self,
        key: str,
        default_value: typing.Union[dict, list],
        evaluation_context: typing.Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[typing.Union[dict, list]]:
        return self._resolve(key, FlagType.OBJECT, default_value, evaluation_context)

    def _resolve(  # noqa: PLR0915 C901
        self,
        flag_key: str,
        flag_type: FlagType,
        default_value: T,
        evaluation_context: typing.Optional[EvaluationContext],
    ) -> FlagResolutionDetails[T]:
        if self.cache is not None and flag_key in self.cache:
            cached_flag: FlagResolutionDetails[T] = self.cache[flag_key]
            cached_flag.reason = Reason.CACHED
            return cached_flag

        context = self._convert_context(evaluation_context)
        call_args = {"timeout": self.deadline, "wait_for_ready": True}
        try:
            request: Message
            if flag_type == FlagType.BOOLEAN:
                request = evaluation_pb2.ResolveBooleanRequest(
                    flag_key=flag_key, context=context
                )
                response = self.stub.ResolveBoolean(request, **call_args)
                value = response.value
            elif flag_type == FlagType.STRING:
                request = evaluation_pb2.ResolveStringRequest(
                    flag_key=flag_key, context=context
                )
                response = self.stub.ResolveString(request, **call_args)
                value = response.value
            elif flag_type == FlagType.OBJECT:
                request = evaluation_pb2.ResolveObjectRequest(
                    flag_key=flag_key, context=context
                )
                response = self.stub.ResolveObject(request, **call_args)
                value = MessageToDict(response, preserving_proto_field_name=True)[
                    "value"
                ]
            elif flag_type == FlagType.FLOAT:
                request = evaluation_pb2.ResolveFloatRequest(
                    flag_key=flag_key, context=context
                )
                response = self.stub.ResolveFloat(request, **call_args)
                value = response.value
            elif flag_type == FlagType.INTEGER:
                request = evaluation_pb2.ResolveIntRequest(
                    flag_key=flag_key, context=context
                )
                response = self.stub.ResolveInt(request, **call_args)
                value = response.value
            else:
                raise ValueError(f"Unknown flag type: {flag_type}")

        except grpc.RpcError as e:
            code = e.code()
            message = f"received grpc status code {code}"

            if code == grpc.StatusCode.NOT_FOUND:
                raise FlagNotFoundError(message) from e
            elif code == grpc.StatusCode.INVALID_ARGUMENT:
                raise TypeMismatchError(message) from e
            elif code == grpc.StatusCode.DATA_LOSS:
                raise ParseError(message) from e
            raise GeneralError(message) from e

        # Got a valid flag and valid type. Return it.
        result = FlagResolutionDetails(
            value=value,
            reason=response.reason,
            variant=response.variant,
        )

        if response.reason == Reason.STATIC and self.cache is not None:
            self.cache.insert(flag_key, result)

        return result

    def _convert_context(
        self, evaluation_context: typing.Optional[EvaluationContext]
    ) -> Struct:
        s = Struct()
        if evaluation_context:
            try:
                s["targetingKey"] = evaluation_context.targeting_key
                s.update(evaluation_context.attributes)
            except ValueError as exc:
                message = (
                    "could not serialize evaluation context to google.protobuf.Struct"
                )
                raise InvalidContextError(message) from exc
        return s



import requests
from asserts import assert_equal
from pytest_bdd import given, parsers, then, when

from openfeature.client import OpenFeatureClient
from openfeature.evaluation_context import EvaluationContext
from openfeature.event import EventDetails
from openfeature.flag_evaluation import FlagEvaluationDetails, Reason

from ._utils import JsonPrimitive, type_cast


@given(
    parsers.cfparse(
        'a {type_info}-flag with key "{key}" and a default value "{default}"'
    ),
    target_fixture="key_and_default_and_type",
)
def setup_key_and_default(
    key: str, default: JsonPrimitive, type_info: str
) -> tuple[str, JsonPrimitive, str]:
    return key, default, type_info


@when("the flag was evaluated with details", target_fixture="details")
def evaluate_with_details(
    client: OpenFeatureClient,
    key_and_default_and_type: tuple,
    evaluation_context: EvaluationContext,
):
    key, default, type_info = key_and_default_and_type
    default = type_cast[type_info](default)
    if type_info == "Boolean":
        return client.get_boolean_details(key, default, evaluation_context)
    elif type_info == "String":
        return client.get_string_details(key, default, evaluation_context)
    elif type_info == "Integer":
        return client.get_integer_details(key, default, evaluation_context)
    elif type_info == "Float":
        return client.get_float_details(key, default, evaluation_context)
    elif type_info == "Object":
        return client.get_object_details(key, default, evaluation_context)
    raise AssertionError("no valid object type")


@when("the flag was modified")
def assert_flag_change_event(container):
    requests.post(f"{container.get_launchpad_url()}/change", timeout=1)


@then("the flag should be part of the event payload")
def assert_flag_change(key_and_default_and_type: tuple, event_details: EventDetails):
    key, _, _ = key_and_default_and_type
    assert key in event_details.flags_changed


@then(
    parsers.cfparse('the resolved details value should be ""'),
)
def resolve_details_value_string(
    details: FlagEvaluationDetails[JsonPrimitive],
    key_and_default_and_type: tuple,
):
    resolve_details_value(details, key_and_default_and_type, "")


@then(
    parsers.cfparse('the resolved details value should be "{value}"'),
)
def resolve_details_value(
    details: FlagEvaluationDetails[JsonPrimitive],
    key_and_default_and_type: tuple,
    value: str,
):
    _, _, type_info = key_and_default_and_type
    assert_equal(details.value, type_cast[type_info](value))


@then(
    parsers.cfparse('the variant should be "{variant}"'),
)
def resolve_details_variant(
    details: FlagEvaluationDetails[JsonPrimitive],
    variant: str,
):
    assert_equal(details.variant, variant)


@then(
    parsers.cfparse('the reason should be "{reason}"'),
)
def resolve_details_reason(
    details: FlagEvaluationDetails[JsonPrimitive],
    reason: str,
):
    assert_equal(details.reason, Reason(reason))



import re
import typing

import pytest
from asserts import assert_equal, assert_true
from pytest_bdd import given, parsers, then, when

from openfeature.contrib.provider.flagd.config import CacheType, Config, ResolverType


def camel_to_snake(name):
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def str2bool(v):
    return v.lower() in ("yes", "true", "t", "1")


def convert_resolver_type(val: typing.Union[str, ResolverType]) -> ResolverType:
    if isinstance(val, str):
        v = val.lower()
        return ResolverType(v)
    else:
        return ResolverType(val)


type_cast = {
    "Integer": int,
    "Long": int,
    "String": str,
    "Boolean": str2bool,
    "ResolverType": convert_resolver_type,
    "CacheType": CacheType,
}


@pytest.fixture()
def option_values() -> dict:
    return {}


@given(
    parsers.cfparse(
        'an option "{option}" of type "{type_info}" with value "{value}"',
    ),
)
def option_with_value(option: str, value: str, type_info: str, option_values: dict):
    value = type_cast[type_info](value)
    option_values[camel_to_snake(option)] = value


@given(
    parsers.cfparse(
        'an environment variable "{env}" with value "{value}"',
    ),
)
def env_with_value(monkeypatch, env: str, value: str):
    monkeypatch.setenv(env, value)


@when(
    parsers.cfparse(
        "a config was initialized",
    ),
    target_fixture="config_or_error",
)
def initialize_config(option_values):
    try:
        return Config(**option_values), False
    except AttributeError:
        return None, True


@when(
    parsers.cfparse(
        'a config was initialized for "{resolver_type}"',
    ),
    target_fixture="config_or_error",
)
def initialize_config_for(resolver_type: str, option_values: dict):
    try:
        return Config(resolver=ResolverType(resolver_type), **option_values), False
    except AttributeError:
        return None, True


@then(
    parsers.cfparse(
        'the option "{option}" of type "{type_info}" should have the value "{value}"',
    )
)
def check_option_value(option, value, type_info, config_or_error):
    value = type_cast[type_info](value)
    value = value if value != "null" else None
    config, _ = config_or_error
    assert_equal(config.__getattribute__(camel_to_snake(option)), value)


@then(
    parsers.cfparse(
        "we should have an error",
    )
)
def check_option_error(config_or_error):
    _, error = config_or_error
    assert_true(error)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: openfeature/schemas/protobuf/sync/v1/sync_service.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'openfeature/schemas/protobuf/sync/v1/sync_service.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n7openfeature/schemas/protobuf/sync/v1/sync_service.proto\x12\x07sync.v1\"9\n\x10SyncFlagsRequest\x12\x13\n\x0bprovider_id\x18\x01 \x01(\t\x12\x10\n\x08selector\x18\x02 \x01(\t\"R\n\x11SyncFlagsResponse\x12\x1a\n\x12\x66lag_configuration\x18\x01 \x01(\t\x12!\n\x05state\x18\x02 \x01(\x0e\x32\x12.sync.v1.SyncState\"=\n\x14\x46\x65tchAllFlagsRequest\x12\x13\n\x0bprovider_id\x18\x01 \x01(\t\x12\x10\n\x08selector\x18\x02 \x01(\t\"3\n\x15\x46\x65tchAllFlagsResponse\x12\x1a\n\x12\x66lag_configuration\x18\x01 \x01(\t*\x92\x01\n\tSyncState\x12\x1a\n\x16SYNC_STATE_UNSPECIFIED\x10\x00\x12\x12\n\x0eSYNC_STATE_ALL\x10\x01\x12\x12\n\x0eSYNC_STATE_ADD\x10\x02\x12\x15\n\x11SYNC_STATE_UPDATE\x10\x03\x12\x15\n\x11SYNC_STATE_DELETE\x10\x04\x12\x13\n\x0fSYNC_STATE_PING\x10\x05\x32\xab\x01\n\x0f\x46lagSyncService\x12\x46\n\tSyncFlags\x12\x19.sync.v1.SyncFlagsRequest\x1a\x1a.sync.v1.SyncFlagsResponse\"\x00\x30\x01\x12P\n\rFetchAllFlags\x12\x1d.sync.v1.FetchAllFlagsRequest\x1a\x1e.sync.v1.FetchAllFlagsResponse\"\x00\x42\x97\x01\n\x1a\x64\x65v.openfeature.flagd.syncZ\x0e\x66lagd/grpcsync\xb8\x01\x01\xaa\x02\x16OpenFeature.Flagd.Sync\xca\x02\'OpenFeature\\Providers\\Flagd\\Schema\\Sync\xea\x02\"OpenFeature::FlagD::Provider::Syncb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'openfeature.schemas.protobuf.sync.v1.sync_service_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\032dev.openfeature.flagd.syncZ\016flagd/grpcsync\270\001\001\252\002\026OpenFeature.Flagd.Sync\312\002\'OpenFeature\\Providers\\Flagd\\Schema\\Sync\352\002\"OpenFeature::FlagD::Provider::Sync'
  _globals['_SYNCSTATE']._serialized_start=328
  _globals['_SYNCSTATE']._serialized_end=474
  _globals['_SYNCFLAGSREQUEST']._serialized_start=68
  _globals['_SYNCFLAGSREQUEST']._serialized_end=125
  _globals['_SYNCFLAGSRESPONSE']._serialized_start=127
  _globals['_SYNCFLAGSRESPONSE']._serialized_end=209
  _globals['_FETCHALLFLAGSREQUEST']._serialized_start=211
  _globals['_FETCHALLFLAGSREQUEST']._serialized_end=272
  _globals['_FETCHALLFLAGSRESPONSE']._serialized_start=274
  _globals['_FETCHALLFLAGSRESPONSE']._serialized_end=325
  _globals['_FLAGSYNCSERVICE']._serialized_start=477
  _globals['_FLAGSYNCSERVICE']._serialized_end=648
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
*
DEPRECATED; use flagd.evaluation.v1
Flag evaluation API

This proto forms the basis of a flag-evaluation API.
It supports single and bulk evaluation RPCs, and flags of various types, as well as establishing a stream for getting notifications about changes in a flag definition.
It supports the inclusion of a "context" with each evaluation, which may contain arbitrary attributes relevant to flag evaluation.
"""

import abc
import collections.abc
import grpc
import grpc.aio
import openfeature.schemas.protobuf.schema.v1.schema_pb2
import typing

_T = typing.TypeVar("_T")

class _MaybeAsyncIterator(collections.abc.AsyncIterator[_T], collections.abc.Iterator[_T], metaclass=abc.ABCMeta): ...

class _ServicerContext(grpc.ServicerContext, grpc.aio.ServicerContext):  # type: ignore[misc, type-arg]
    ...

class ServiceStub:
    """Service defines the exposed rpcs of flagd"""

    def __init__(self, channel: typing.Union[grpc.Channel, grpc.aio.Channel]) -> None: ...
    ResolveAll: grpc.UnaryUnaryMultiCallable[
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveAllRequest,
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveAllResponse,
    ]

    ResolveBoolean: grpc.UnaryUnaryMultiCallable[
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveBooleanRequest,
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveBooleanResponse,
    ]

    ResolveString: grpc.UnaryUnaryMultiCallable[
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveStringRequest,
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveStringResponse,
    ]

    ResolveFloat: grpc.UnaryUnaryMultiCallable[
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveFloatRequest,
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveFloatResponse,
    ]

    ResolveInt: grpc.UnaryUnaryMultiCallable[
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveIntRequest,
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveIntResponse,
    ]

    ResolveObject: grpc.UnaryUnaryMultiCallable[
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveObjectRequest,
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveObjectResponse,
    ]

    EventStream: grpc.UnaryStreamMultiCallable[
        openfeature.schemas.protobuf.schema.v1.schema_pb2.EventStreamRequest,
        openfeature.schemas.protobuf.schema.v1.schema_pb2.EventStreamResponse,
    ]

class ServiceAsyncStub:
    """Service defines the exposed rpcs of flagd"""

    ResolveAll: grpc.aio.UnaryUnaryMultiCallable[
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveAllRequest,
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveAllResponse,
    ]

    ResolveBoolean: grpc.aio.UnaryUnaryMultiCallable[
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveBooleanRequest,
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveBooleanResponse,
    ]

    ResolveString: grpc.aio.UnaryUnaryMultiCallable[
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveStringRequest,
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveStringResponse,
    ]

    ResolveFloat: grpc.aio.UnaryUnaryMultiCallable[
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveFloatRequest,
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveFloatResponse,
    ]

    ResolveInt: grpc.aio.UnaryUnaryMultiCallable[
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveIntRequest,
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveIntResponse,
    ]

    ResolveObject: grpc.aio.UnaryUnaryMultiCallable[
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveObjectRequest,
        openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveObjectResponse,
    ]

    EventStream: grpc.aio.UnaryStreamMultiCallable[
        openfeature.schemas.protobuf.schema.v1.schema_pb2.EventStreamRequest,
        openfeature.schemas.protobuf.schema.v1.schema_pb2.EventStreamResponse,
    ]

class ServiceServicer(metaclass=abc.ABCMeta):
    """Service defines the exposed rpcs of flagd"""

    @abc.abstractmethod
    def ResolveAll(
        self,
        request: openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveAllRequest,
        context: _ServicerContext,
    ) -> typing.Union[openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveAllResponse, collections.abc.Awaitable[openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveAllResponse]]: ...

    @abc.abstractmethod
    def ResolveBoolean(
        self,
        request: openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveBooleanRequest,
        context: _ServicerContext,
    ) -> typing.Union[openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveBooleanResponse, collections.abc.Awaitable[openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveBooleanResponse]]: ...

    @abc.abstractmethod
    def ResolveString(
        self,
        request: openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveStringRequest,
        context: _ServicerContext,
    ) -> typing.Union[openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveStringResponse, collections.abc.Awaitable[openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveStringResponse]]: ...

    @abc.abstractmethod
    def ResolveFloat(
        self,
        request: openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveFloatRequest,
        context: _ServicerContext,
    ) -> typing.Union[openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveFloatResponse, collections.abc.Awaitable[openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveFloatResponse]]: ...

    @abc.abstractmethod
    def ResolveInt(
        self,
        request: openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveIntRequest,
        context: _ServicerContext,
    ) -> typing.Union[openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveIntResponse, collections.abc.Awaitable[openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveIntResponse]]: ...

    @abc.abstractmethod
    def ResolveObject(
        self,
        request: openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveObjectRequest,
        context: _ServicerContext,
    ) -> typing.Union[openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveObjectResponse, collections.abc.Awaitable[openfeature.schemas.protobuf.schema.v1.schema_pb2.ResolveObjectResponse]]: ...

    @abc.abstractmethod
    def EventStream(
        self,
        request: openfeature.schemas.protobuf.schema.v1.schema_pb2.EventStreamRequest,
        context: _ServicerContext,
    ) -> typing.Union[collections.abc.Iterator[openfeature.schemas.protobuf.schema.v1.schema_pb2.EventStreamResponse], collections.abc.AsyncIterator[openfeature.schemas.protobuf.schema.v1.schema_pb2.EventStreamResponse]]: ...

def add_ServiceServicer_to_server(servicer: ServiceServicer, server: typing.Union[grpc.Server, grpc.aio.Server]) -> None: ...



import os.path
import time
import typing
from pathlib import Path

import grpc
from grpc_health.v1 import health_pb2, health_pb2_grpc
from testcontainers.core.container import DockerContainer
from testcontainers.core.waiting_utils import wait_container_is_ready, wait_for_logs

from openfeature.contrib.provider.flagd.config import ResolverType

HEALTH_CHECK = 8014
LAUNCHPAD = 8080


class FlagdContainer(DockerContainer):
    def __init__(
        self,
        feature: typing.Optional[str] = None,
        **kwargs,
    ) -> None:
        image: str = "ghcr.io/open-feature/flagd-testbed"
        if feature is not None:
            image = f"{image}-{feature}"
        path = Path(__file__).parents[2] / "openfeature/test-harness/version.txt"
        data = path.read_text().rstrip()
        super().__init__(f"{image}:v{data}", **kwargs)
        self.rpc = 8013
        self.ipr = 8015
        self.flagDir = Path("./flags")
        self.flagDir.mkdir(parents=True, exist_ok=True)
        self.with_exposed_ports(self.rpc, self.ipr, HEALTH_CHECK, LAUNCHPAD)
        self.with_volume_mapping(os.path.abspath(self.flagDir.name), "/flags", "rw")

    def get_port(self, resolver_type: ResolverType):
        if resolver_type == ResolverType.RPC:
            return self.get_exposed_port(self.rpc)
        else:
            return self.get_exposed_port(self.ipr)

    def get_launchpad_url(self):
        return f"http://localhost:{self.get_exposed_port(LAUNCHPAD)}"

    def start(self) -> "FlagdContainer":
        super().start()
        self._checker(self.get_container_host_ip(), self.get_exposed_port(HEALTH_CHECK))
        return self

    @wait_container_is_ready(ConnectionError)
    def _checker(self, host: str, port: str) -> None:
        # First we wait for Flagd to say it's listening
        wait_for_logs(
            self,
            "listening",
            5,
        )

        time.sleep(1)
        # Second we use the GRPC health check endpoint
        with grpc.insecure_channel(host + ":" + port) as channel:
            health_stub = health_pb2_grpc.HealthStub(channel)

            def health_check_call(stub: health_pb2_grpc.HealthStub):
                request = health_pb2.HealthCheckRequest()
                resp = stub.Check(request)
                if resp.status == health_pb2.HealthCheckResponse.SERVING:
                    return True
                elif resp.status == health_pb2.HealthCheckResponse.NOT_SERVING:
                    return False

            # Should succeed
            # Check health status every 1 second for 30 seconds
            ok = False
            for _ in range(30):
                ok = health_check_call(health_stub)
                if ok:
                    break
                time.sleep(1)

            if not ok:
                raise ConnectionError("flagD not ready in time")



import json
import logging
import threading
import time
import typing

import grpc
from google.protobuf.json_format import MessageToDict

from openfeature.evaluation_context import EvaluationContext
from openfeature.event import ProviderEventDetails
from openfeature.exception import ErrorCode, ParseError, ProviderNotReadyError
from openfeature.schemas.protobuf.flagd.sync.v1 import (
    sync_pb2,
    sync_pb2_grpc,
)

from ....config import Config
from ..connector import FlagStateConnector
from ..flags import FlagStore

logger = logging.getLogger("openfeature.contrib")


class GrpcWatcher(FlagStateConnector):
    def __init__(
        self,
        config: Config,
        flag_store: FlagStore,
        emit_provider_ready: typing.Callable[[ProviderEventDetails, dict], None],
        emit_provider_error: typing.Callable[[ProviderEventDetails], None],
        emit_provider_stale: typing.Callable[[ProviderEventDetails], None],
    ):
        self.flag_store = flag_store
        self.config = config

        self.channel = self._generate_channel(config)
        self.stub = sync_pb2_grpc.FlagSyncServiceStub(self.channel)
        self.retry_backoff_seconds = config.retry_backoff_ms * 0.001
        self.retry_backoff_max_seconds = config.retry_backoff_ms * 0.001
        self.retry_grace_period = config.retry_grace_period
        self.streamline_deadline_seconds = config.stream_deadline_ms * 0.001
        self.deadline = config.deadline_ms * 0.001
        self.selector = config.selector
        self.provider_id = config.provider_id
        self.emit_provider_ready = emit_provider_ready
        self.emit_provider_error = emit_provider_error
        self.emit_provider_stale = emit_provider_stale

        self.connected = False
        self.thread: typing.Optional[threading.Thread] = None
        self.timer: typing.Optional[threading.Timer] = None

        self.start_time = time.time()

    def _generate_channel(self, config: Config) -> grpc.Channel:
        target = f"{config.host}:{config.port}"
        # Create the channel with the service config
        options: list[tuple[str, typing.Any]] = [
            ("grpc.keepalive_time_ms", config.keep_alive_time),
            ("grpc.initial_reconnect_backoff_ms", config.retry_backoff_ms),
            ("grpc.max_reconnect_backoff_ms", config.retry_backoff_max_ms),
            ("grpc.min_reconnect_backoff_ms", config.stream_deadline_ms),
        ]
        if config.default_authority is not None:
            options.append(("grpc.default_authority", config.default_authority))

        if config.channel_credentials is not None:
            channel_args = {
                "options": options,
                "credentials": config.channel_credentials,
            }
            channel = grpc.secure_channel(target, **channel_args)

        elif config.tls:
            channel_args = {
                "options": options,
                "credentials": grpc.ssl_channel_credentials(),
            }
            if config.cert_path:
                with open(config.cert_path, "rb") as f:
                    channel_args["credentials"] = grpc.ssl_channel_credentials(f.read())

            channel = grpc.secure_channel(target, **channel_args)

        else:
            channel = grpc.insecure_channel(
                target,
                options=options,
            )

        return channel

    def initialize(self, context: EvaluationContext) -> None:
        self.connect()

    def connect(self) -> None:
        self.active = True

        # Run monitoring in a separate thread
        self.monitor_thread = threading.Thread(
            target=self.monitor, daemon=True, name="FlagdGrpcSyncServiceMonitorThread"
        )
        self.monitor_thread.start()
        ## block until ready or deadline reached
        timeout = self.deadline + time.time()
        while not self.connected and time.time() < timeout:
            time.sleep(0.05)
        logger.debug("Finished blocking gRPC state initialization")

        if not self.connected:
            raise ProviderNotReadyError(
                "Blocking init finished before data synced. Consider increasing startup deadline to avoid inconsistent evaluations."
            )

    def monitor(self) -> None:
        self.channel.subscribe(self._state_change_callback, try_to_connect=True)

    def _state_change_callback(self, new_state: grpc.ChannelConnectivity) -> None:
        logger.debug(f"gRPC state change: {new_state}")
        if (
            new_state == grpc.ChannelConnectivity.READY
            or new_state == grpc.ChannelConnectivity.IDLE
        ):
            if not self.thread or not self.thread.is_alive():
                self.thread = threading.Thread(
                    target=self.listen,
                    daemon=True,
                    name="FlagdGrpcSyncWorkerThread",
                )
                self.thread.start()

            if self.timer and self.timer.is_alive():
                logger.debug("gRPC error timer expired")
                self.timer.cancel()

        elif new_state == grpc.ChannelConnectivity.TRANSIENT_FAILURE:
            # this is the failed reconnect attempt so we are going into stale
            self.emit_provider_stale(
                ProviderEventDetails(
                    message="gRPC sync disconnected, reconnecting",
                )
            )
            self.start_time = time.time()
            # adding a timer, so we can emit the error event after time
            self.timer = threading.Timer(self.retry_grace_period, self.emit_error)

            logger.debug("gRPC error timer started")
            self.timer.start()
            self.connected = False

    def emit_error(self) -> None:
        logger.debug("gRPC error emitted")
        self.emit_provider_error(
            ProviderEventDetails(
                message="gRPC sync disconnected, reconnecting",
                error_code=ErrorCode.GENERAL,
            )
        )

    def shutdown(self) -> None:
        self.active = False
        self.channel.close()

    def listen(self) -> None:
        call_args = (
            {"timeout": self.streamline_deadline_seconds}
            if self.streamline_deadline_seconds > 0
            else {}
        )
        request_args = {}
        if self.selector is not None:
            request_args["selector"] = self.selector
        if self.provider_id is not None:
            request_args["provider_id"] = self.provider_id

        while self.active:
            try:
                context_values_request = sync_pb2.GetMetadataRequest()
                context_values_response: sync_pb2.GetMetadataResponse = (
                    self.stub.GetMetadata(context_values_request, wait_for_ready=True)
                )
                context_values = MessageToDict(context_values_response)

                request = sync_pb2.SyncFlagsRequest(**request_args)

                logger.debug("Setting up gRPC sync flags connection")
                for flag_rsp in self.stub.SyncFlags(
                    request, wait_for_ready=True, **call_args
                ):
                    flag_str = flag_rsp.flag_configuration
                    logger.debug(
                        f"Received flag configuration - {abs(hash(flag_str)) % (10**8)}"
                    )
                    self.flag_store.update(json.loads(flag_str))

                    if not self.connected:
                        self.emit_provider_ready(
                            ProviderEventDetails(
                                message="gRPC sync connection established"
                            ),
                            context_values["metadata"],
                        )
                        self.connected = True

                    if not self.active:
                        logger.debug("Terminating gRPC sync thread")
                        return
            except grpc.RpcError as e:  # noqa: PERF203
                logger.error(f"SyncFlags stream error, {e.code()=} {e.details()=}")
            except json.JSONDecodeError:
                logger.exception(
                    f"Could not parse JSON flag data from SyncFlags endpoint: {flag_str=}"
                )
            except ParseError:
                logger.exception(
                    f"Could not parse flag data using flagd syntax: {flag_str=}"
                )



import logging
import time

import pytest
from asserts import assert_greater
from pytest_bdd import given, parsers, then, when

from openfeature.client import OpenFeatureClient
from openfeature.event import ProviderEvent

events = {
    "ready": ProviderEvent.PROVIDER_READY,
    "error": ProviderEvent.PROVIDER_ERROR,
    "stale": ProviderEvent.PROVIDER_STALE,
    "change": ProviderEvent.PROVIDER_CONFIGURATION_CHANGED,
}


@pytest.fixture()
def event_handles() -> list:
    return []


@given(
    parsers.cfparse(
        "a {event_type} event handler",
    ),
)
def add_event_handler(client: OpenFeatureClient, event_type: str, event_handles: list):
    def handler(event):
        logging.warning((event_type, event))
        event_handles.append(
            {
                "type": event_type,
                "event": event,
            }
        )

    client.add_handler(events[event_type], handler)

    logging.warning(("handler added", event_type))


def assert_handlers(handles, event_type: str, max_wait: int = 2):
    poll_interval = 1
    while max_wait > 0:
        found = any(h["type"] == event_type for h in handles)
        if not found:
            max_wait -= poll_interval
            time.sleep(poll_interval)
            continue
        break
    return handles


@when(
    parsers.cfparse(
        "a {event_type} event was fired",
    ),
    target_fixture="event_details",
)
def pass_for_event_fired(event_type: str, event_handles):
    events = assert_handlers(event_handles, event_type, 30000)
    events = [e for e in events if e["type"] == event_type]
    assert_greater(len(events), 0)
    for event in event_handles:
        event_handles.remove(event)
    return events[0]["event"]


@then(
    parsers.cfparse(
        "the {event_type} event handler should have been executed",
    )
)
def assert_handler_run(event_type, event_handles):
    assert_handler_run_within(event_type, event_handles, 30000)


@then(
    parsers.cfparse(
        "the {event_type} event handler should have been executed within {time:d}ms",
    )
)
def assert_handler_run_within(event_type, event_handles, time: int):
    events = assert_handlers(event_handles, event_type, max_wait=int(time / 1000))
    assert_greater(len(events), 0)

    for event in event_handles:
        event_handles.remove(event)



import typing

from openfeature.evaluation_context import EvaluationContext


class FlagStateConnector(typing.Protocol):
    def initialize(
        self, evaluation_context: EvaluationContext
    ) -> None: ...  # pragma: no cover

    def shutdown(self) -> None: ...  # pragma: no cover



import dataclasses
import os
import typing
from enum import Enum

import grpc


class ResolverType(Enum):
    RPC = "rpc"
    IN_PROCESS = "in-process"
    FILE = "file"


class CacheType(Enum):
    LRU = "lru"
    DISABLED = "disabled"


DEFAULT_CACHE = CacheType.LRU
DEFAULT_CACHE_SIZE = 1000
DEFAULT_DEADLINE = 500
DEFAULT_HOST = "localhost"
DEFAULT_KEEP_ALIVE = 0
DEFAULT_OFFLINE_SOURCE_PATH: typing.Optional[str] = None
DEFAULT_OFFLINE_POLL_MS = 5000
DEFAULT_PORT_IN_PROCESS = 8015
DEFAULT_PORT_RPC = 8013
DEFAULT_RESOLVER_TYPE = ResolverType.RPC
DEFAULT_RETRY_BACKOFF = 1000
DEFAULT_RETRY_BACKOFF_MAX = 120000
DEFAULT_RETRY_GRACE_PERIOD_SECONDS = 5
DEFAULT_STREAM_DEADLINE = 600000
DEFAULT_TLS = False
DEFAULT_TLS_CERT: typing.Optional[str] = None

ENV_VAR_CACHE_SIZE = "FLAGD_MAX_CACHE_SIZE"
ENV_VAR_CACHE_TYPE = "FLAGD_CACHE"
ENV_VAR_DEADLINE_MS = "FLAGD_DEADLINE_MS"
ENV_VAR_HOST = "FLAGD_HOST"
ENV_VAR_KEEP_ALIVE_TIME_MS = "FLAGD_KEEP_ALIVE_TIME_MS"
ENV_VAR_OFFLINE_FLAG_SOURCE_PATH = "FLAGD_OFFLINE_FLAG_SOURCE_PATH"
ENV_VAR_OFFLINE_POLL_MS = "FLAGD_OFFLINE_POLL_MS"
ENV_VAR_PORT = "FLAGD_PORT"
ENV_VAR_RESOLVER_TYPE = "FLAGD_RESOLVER"
ENV_VAR_RETRY_BACKOFF_MS = "FLAGD_RETRY_BACKOFF_MS"
ENV_VAR_RETRY_BACKOFF_MAX_MS = "FLAGD_RETRY_BACKOFF_MAX_MS"
ENV_VAR_RETRY_GRACE_PERIOD_SECONDS = "FLAGD_RETRY_GRACE_PERIOD"
ENV_VAR_SELECTOR = "FLAGD_SOURCE_SELECTOR"
ENV_VAR_PROVIDER_ID = "FLAGD_SOURCE_PROVIDER_ID"
ENV_VAR_STREAM_DEADLINE_MS = "FLAGD_STREAM_DEADLINE_MS"
ENV_VAR_TLS = "FLAGD_TLS"
ENV_VAR_TLS_CERT = "FLAGD_SERVER_CERT_PATH"
ENV_VAR_DEFAULT_AUTHORITY = "FLAGD_DEFAULT_AUTHORITY"

T = typing.TypeVar("T")


def str_to_bool(val: str) -> bool:
    return val.lower() == "true"


def convert_resolver_type(val: typing.Union[str, ResolverType]) -> ResolverType:
    if isinstance(val, str):
        v = val.lower()
        return ResolverType(v)
    else:
        return ResolverType(val)


def env_or_default(
    env_var: str, default: T, cast: typing.Optional[typing.Callable[[str], T]] = None
) -> typing.Union[str, T]:
    val = os.environ.get(env_var)
    if val is None:
        return default
    return val if cast is None else cast(val)


@dataclasses.dataclass
class Config:
    def __init__(  # noqa: PLR0913
        self,
        host: typing.Optional[str] = None,
        port: typing.Optional[int] = None,
        tls: typing.Optional[bool] = None,
        selector: typing.Optional[str] = None,
        provider_id: typing.Optional[str] = None,
        resolver: typing.Optional[ResolverType] = None,
        offline_flag_source_path: typing.Optional[str] = None,
        offline_poll_interval_ms: typing.Optional[int] = None,
        retry_backoff_ms: typing.Optional[int] = None,
        retry_backoff_max_ms: typing.Optional[int] = None,
        retry_grace_period: typing.Optional[int] = None,
        deadline_ms: typing.Optional[int] = None,
        stream_deadline_ms: typing.Optional[int] = None,
        keep_alive_time: typing.Optional[int] = None,
        cache: typing.Optional[CacheType] = None,
        max_cache_size: typing.Optional[int] = None,
        cert_path: typing.Optional[str] = None,
        default_authority: typing.Optional[str] = None,
        channel_credentials: typing.Optional[grpc.ChannelCredentials] = None,
    ):
        self.host = env_or_default(ENV_VAR_HOST, DEFAULT_HOST) if host is None else host

        self.tls = (
            env_or_default(ENV_VAR_TLS, DEFAULT_TLS, cast=str_to_bool)
            if tls is None
            else tls
        )

        self.retry_backoff_ms: int = (
            int(
                env_or_default(
                    ENV_VAR_RETRY_BACKOFF_MS, DEFAULT_RETRY_BACKOFF, cast=int
                )
            )
            if retry_backoff_ms is None
            else retry_backoff_ms
        )
        self.retry_backoff_max_ms: int = (
            int(
                env_or_default(
                    ENV_VAR_RETRY_BACKOFF_MAX_MS, DEFAULT_RETRY_BACKOFF_MAX, cast=int
                )
            )
            if retry_backoff_max_ms is None
            else retry_backoff_max_ms
        )

        self.retry_grace_period: int = (
            int(
                env_or_default(
                    ENV_VAR_RETRY_GRACE_PERIOD_SECONDS,
                    DEFAULT_RETRY_GRACE_PERIOD_SECONDS,
                    cast=int,
                )
            )
            if retry_grace_period is None
            else retry_grace_period
        )

        self.resolver = (
            env_or_default(
                ENV_VAR_RESOLVER_TYPE, DEFAULT_RESOLVER_TYPE, cast=convert_resolver_type
            )
            if resolver is None
            else resolver
        )

        default_port = (
            DEFAULT_PORT_RPC
            if self.resolver is ResolverType.RPC
            else DEFAULT_PORT_IN_PROCESS
        )

        self.port: int = (
            int(env_or_default(ENV_VAR_PORT, default_port, cast=int))
            if port is None
            else port
        )

        self.offline_flag_source_path = (
            env_or_default(
                ENV_VAR_OFFLINE_FLAG_SOURCE_PATH, DEFAULT_OFFLINE_SOURCE_PATH
            )
            if offline_flag_source_path is None
            else offline_flag_source_path
        )

        if (
            self.offline_flag_source_path is not None
            and self.resolver is ResolverType.IN_PROCESS
        ):
            self.resolver = ResolverType.FILE

        if self.resolver is ResolverType.FILE and self.offline_flag_source_path is None:
            raise AttributeError(
                "Resolver Type 'FILE' requires a offlineFlagSourcePath"
            )

        self.offline_poll_interval_ms: int = (
            int(
                env_or_default(
                    ENV_VAR_OFFLINE_POLL_MS, DEFAULT_OFFLINE_POLL_MS, cast=int
                )
            )
            if offline_poll_interval_ms is None
            else offline_poll_interval_ms
        )

        self.deadline_ms: int = (
            int(env_or_default(ENV_VAR_DEADLINE_MS, DEFAULT_DEADLINE, cast=int))
            if deadline_ms is None
            else deadline_ms
        )

        self.stream_deadline_ms: int = (
            int(
                env_or_default(
                    ENV_VAR_STREAM_DEADLINE_MS, DEFAULT_STREAM_DEADLINE, cast=int
                )
            )
            if stream_deadline_ms is None
            else stream_deadline_ms
        )

        self.keep_alive_time: int = (
            int(
                env_or_default(ENV_VAR_KEEP_ALIVE_TIME_MS, DEFAULT_KEEP_ALIVE, cast=int)
            )
            if keep_alive_time is None
            else keep_alive_time
        )

        self.cache = (
            CacheType(env_or_default(ENV_VAR_CACHE_TYPE, DEFAULT_CACHE))
            if cache is None
            else cache
        )

        self.max_cache_size: int = (
            int(env_or_default(ENV_VAR_CACHE_SIZE, DEFAULT_CACHE_SIZE, cast=int))
            if max_cache_size is None
            else max_cache_size
        )

        self.cert_path = (
            env_or_default(ENV_VAR_TLS_CERT, DEFAULT_TLS_CERT)
            if cert_path is None
            else cert_path
        )

        self.selector = (
            env_or_default(ENV_VAR_SELECTOR, None) if selector is None else selector
        )

        self.provider_id = (
            env_or_default(ENV_VAR_PROVIDER_ID, None)
            if provider_id is None
            else provider_id
        )

        self.default_authority = (
            env_or_default(ENV_VAR_DEFAULT_AUTHORITY, None)
            if default_authority is None
            else default_authority
        )

        self.channel_credentials = channel_credentials



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
*
Flag evaluation API

This proto forms the basis of a flag-evaluation API.
It supports single and bulk evaluation RPCs, and flags of various types, as well as establishing a stream for getting notifications about changes in a flag definition.
It supports the inclusion of a "context" with each evaluation, which may contain arbitrary attributes relevant to flag evaluation.
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import google.protobuf.struct_pb2
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class ResolveAllRequest(google.protobuf.message.Message):
    """Request body for bulk flag evaluation, used by the ResolveAll rpc."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CONTEXT_FIELD_NUMBER: builtins.int
    @property
    def context(self) -> google.protobuf.struct_pb2.Struct:
        """Object structure describing the EvaluationContext used in the flag evaluation, see https://openfeature.dev/docs/reference/concepts/evaluation-context"""

    def __init__(
        self,
        *,
        context: google.protobuf.struct_pb2.Struct | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["context", b"context"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["context", b"context"]) -> None: ...

global___ResolveAllRequest = ResolveAllRequest

@typing.final
class ResolveAllResponse(google.protobuf.message.Message):
    """Response body for bulk flag evaluation, used by the ResolveAll rpc."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class FlagsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___AnyFlag: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: global___AnyFlag | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    FLAGS_FIELD_NUMBER: builtins.int
    @property
    def flags(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___AnyFlag]:
        """Object structure describing the evaluated flags for the provided context."""

    def __init__(
        self,
        *,
        flags: collections.abc.Mapping[builtins.str, global___AnyFlag] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["flags", b"flags"]) -> None: ...

global___ResolveAllResponse = ResolveAllResponse

@typing.final
class AnyFlag(google.protobuf.message.Message):
    """A variant type flag response."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REASON_FIELD_NUMBER: builtins.int
    VARIANT_FIELD_NUMBER: builtins.int
    BOOL_VALUE_FIELD_NUMBER: builtins.int
    STRING_VALUE_FIELD_NUMBER: builtins.int
    DOUBLE_VALUE_FIELD_NUMBER: builtins.int
    OBJECT_VALUE_FIELD_NUMBER: builtins.int
    reason: builtins.str
    """The reason for the given return value, see https://openfeature.dev/docs/specification/types#resolution-details"""
    variant: builtins.str
    """The variant name of the returned flag value."""
    bool_value: builtins.bool
    string_value: builtins.str
    double_value: builtins.float
    @property
    def object_value(self) -> google.protobuf.struct_pb2.Struct: ...
    def __init__(
        self,
        *,
        reason: builtins.str = ...,
        variant: builtins.str = ...,
        bool_value: builtins.bool = ...,
        string_value: builtins.str = ...,
        double_value: builtins.float = ...,
        object_value: google.protobuf.struct_pb2.Struct | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["bool_value", b"bool_value", "double_value", b"double_value", "object_value", b"object_value", "string_value", b"string_value", "value", b"value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["bool_value", b"bool_value", "double_value", b"double_value", "object_value", b"object_value", "reason", b"reason", "string_value", b"string_value", "value", b"value", "variant", b"variant"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["value", b"value"]) -> typing.Literal["bool_value", "string_value", "double_value", "object_value"] | None: ...

global___AnyFlag = AnyFlag

@typing.final
class ResolveBooleanRequest(google.protobuf.message.Message):
    """Request body for boolean flag evaluation, used by the ResolveBoolean rpc."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FLAG_KEY_FIELD_NUMBER: builtins.int
    CONTEXT_FIELD_NUMBER: builtins.int
    flag_key: builtins.str
    """Flag key of the requested flag."""
    @property
    def context(self) -> google.protobuf.struct_pb2.Struct:
        """Object structure describing the EvaluationContext used in the flag evaluation, see https://openfeature.dev/docs/reference/concepts/evaluation-context"""

    def __init__(
        self,
        *,
        flag_key: builtins.str = ...,
        context: google.protobuf.struct_pb2.Struct | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["context", b"context"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["context", b"context", "flag_key", b"flag_key"]) -> None: ...

global___ResolveBooleanRequest = ResolveBooleanRequest

@typing.final
class ResolveBooleanResponse(google.protobuf.message.Message):
    """Response body for boolean flag evaluation. used by the ResolveBoolean rpc."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUE_FIELD_NUMBER: builtins.int
    REASON_FIELD_NUMBER: builtins.int
    VARIANT_FIELD_NUMBER: builtins.int
    METADATA_FIELD_NUMBER: builtins.int
    value: builtins.bool
    """The response value of the boolean flag evaluation, will be unset in the case of error."""
    reason: builtins.str
    """The reason for the given return value, see https://openfeature.dev/docs/specification/types#resolution-details"""
    variant: builtins.str
    """The variant name of the returned flag value."""
    @property
    def metadata(self) -> google.protobuf.struct_pb2.Struct:
        """Metadata for this evaluation"""

    def __init__(
        self,
        *,
        value: builtins.bool = ...,
        reason: builtins.str = ...,
        variant: builtins.str = ...,
        metadata: google.protobuf.struct_pb2.Struct | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["metadata", b"metadata"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["metadata", b"metadata", "reason", b"reason", "value", b"value", "variant", b"variant"]) -> None: ...

global___ResolveBooleanResponse = ResolveBooleanResponse

@typing.final
class ResolveStringRequest(google.protobuf.message.Message):
    """Request body for string flag evaluation, used by the ResolveString rpc."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FLAG_KEY_FIELD_NUMBER: builtins.int
    CONTEXT_FIELD_NUMBER: builtins.int
    flag_key: builtins.str
    """Flag key of the requested flag."""
    @property
    def context(self) -> google.protobuf.struct_pb2.Struct:
        """Object structure describing the EvaluationContext used in the flag evaluation, see https://openfeature.dev/docs/reference/concepts/evaluation-context"""

    def __init__(
        self,
        *,
        flag_key: builtins.str = ...,
        context: google.protobuf.struct_pb2.Struct | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["context", b"context"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["context", b"context", "flag_key", b"flag_key"]) -> None: ...

global___ResolveStringRequest = ResolveStringRequest

@typing.final
class ResolveStringResponse(google.protobuf.message.Message):
    """Response body for string flag evaluation. used by the ResolveString rpc."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUE_FIELD_NUMBER: builtins.int
    REASON_FIELD_NUMBER: builtins.int
    VARIANT_FIELD_NUMBER: builtins.int
    METADATA_FIELD_NUMBER: builtins.int
    value: builtins.str
    """The response value of the string flag evaluation, will be unset in the case of error."""
    reason: builtins.str
    """The reason for the given return value, see https://openfeature.dev/docs/specification/types#resolution-details"""
    variant: builtins.str
    """The variant name of the returned flag value."""
    @property
    def metadata(self) -> google.protobuf.struct_pb2.Struct:
        """Metadata for this evaluation"""

    def __init__(
        self,
        *,
        value: builtins.str = ...,
        reason: builtins.str = ...,
        variant: builtins.str = ...,
        metadata: google.protobuf.struct_pb2.Struct | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["metadata", b"metadata"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["metadata", b"metadata", "reason", b"reason", "value", b"value", "variant", b"variant"]) -> None: ...

global___ResolveStringResponse = ResolveStringResponse

@typing.final
class ResolveFloatRequest(google.protobuf.message.Message):
    """Request body for float flag evaluation, used by the ResolveFloat rpc."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FLAG_KEY_FIELD_NUMBER: builtins.int
    CONTEXT_FIELD_NUMBER: builtins.int
    flag_key: builtins.str
    """Flag key of the requested flag."""
    @property
    def context(self) -> google.protobuf.struct_pb2.Struct:
        """Object structure describing the EvaluationContext used in the flag evaluation, see https://openfeature.dev/docs/reference/concepts/evaluation-context"""

    def __init__(
        self,
        *,
        flag_key: builtins.str = ...,
        context: google.protobuf.struct_pb2.Struct | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["context", b"context"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["context", b"context", "flag_key", b"flag_key"]) -> None: ...

global___ResolveFloatRequest = ResolveFloatRequest

@typing.final
class ResolveFloatResponse(google.protobuf.message.Message):
    """Response body for float flag evaluation. used by the ResolveFloat rpc."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUE_FIELD_NUMBER: builtins.int
    REASON_FIELD_NUMBER: builtins.int
    VARIANT_FIELD_NUMBER: builtins.int
    METADATA_FIELD_NUMBER: builtins.int
    value: builtins.float
    """The response value of the float flag evaluation, will be empty in the case of error."""
    reason: builtins.str
    """The reason for the given return value, see https://openfeature.dev/docs/specification/types#resolution-details"""
    variant: builtins.str
    """The variant name of the returned flag value."""
    @property
    def metadata(self) -> google.protobuf.struct_pb2.Struct:
        """Metadata for this evaluation"""

    def __init__(
        self,
        *,
        value: builtins.float = ...,
        reason: builtins.str = ...,
        variant: builtins.str = ...,
        metadata: google.protobuf.struct_pb2.Struct | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["metadata", b"metadata"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["metadata", b"metadata", "reason", b"reason", "value", b"value", "variant", b"variant"]) -> None: ...

global___ResolveFloatResponse = ResolveFloatResponse

@typing.final
class ResolveIntRequest(google.protobuf.message.Message):
    """Request body for int flag evaluation, used by the ResolveInt rpc."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FLAG_KEY_FIELD_NUMBER: builtins.int
    CONTEXT_FIELD_NUMBER: builtins.int
    flag_key: builtins.str
    """Flag key of the requested flag."""
    @property
    def context(self) -> google.protobuf.struct_pb2.Struct:
        """Object structure describing the EvaluationContext used in the flag evaluation, see https://openfeature.dev/docs/reference/concepts/evaluation-context"""

    def __init__(
        self,
        *,
        flag_key: builtins.str = ...,
        context: google.protobuf.struct_pb2.Struct | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["context", b"context"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["context", b"context", "flag_key", b"flag_key"]) -> None: ...

global___ResolveIntRequest = ResolveIntRequest

@typing.final
class ResolveIntResponse(google.protobuf.message.Message):
    """Response body for int flag evaluation. used by the ResolveInt rpc."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUE_FIELD_NUMBER: builtins.int
    REASON_FIELD_NUMBER: builtins.int
    VARIANT_FIELD_NUMBER: builtins.int
    METADATA_FIELD_NUMBER: builtins.int
    value: builtins.int
    """The response value of the int flag evaluation, will be unset in the case of error."""
    reason: builtins.str
    """The reason for the given return value, see https://openfeature.dev/docs/specification/types#resolution-details"""
    variant: builtins.str
    """The variant name of the returned flag value."""
    @property
    def metadata(self) -> google.protobuf.struct_pb2.Struct:
        """Metadata for this evaluation"""

    def __init__(
        self,
        *,
        value: builtins.int = ...,
        reason: builtins.str = ...,
        variant: builtins.str = ...,
        metadata: google.protobuf.struct_pb2.Struct | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["metadata", b"metadata"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["metadata", b"metadata", "reason", b"reason", "value", b"value", "variant", b"variant"]) -> None: ...

global___ResolveIntResponse = ResolveIntResponse

@typing.final
class ResolveObjectRequest(google.protobuf.message.Message):
    """Request body for object flag evaluation, used by the ResolveObject rpc."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FLAG_KEY_FIELD_NUMBER: builtins.int
    CONTEXT_FIELD_NUMBER: builtins.int
    flag_key: builtins.str
    """Flag key of the requested flag."""
    @property
    def context(self) -> google.protobuf.struct_pb2.Struct:
        """Object structure describing the EvaluationContext used in the flag evaluation, see https://openfeature.dev/docs/reference/concepts/evaluation-context"""

    def __init__(
        self,
        *,
        flag_key: builtins.str = ...,
        context: google.protobuf.struct_pb2.Struct | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["context", b"context"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["context", b"context", "flag_key", b"flag_key"]) -> None: ...

global___ResolveObjectRequest = ResolveObjectRequest

@typing.final
class ResolveObjectResponse(google.protobuf.message.Message):
    """Response body for object flag evaluation. used by the ResolveObject rpc."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUE_FIELD_NUMBER: builtins.int
    REASON_FIELD_NUMBER: builtins.int
    VARIANT_FIELD_NUMBER: builtins.int
    METADATA_FIELD_NUMBER: builtins.int
    reason: builtins.str
    """The reason for the given return value, see https://openfeature.dev/docs/specification/types#resolution-details"""
    variant: builtins.str
    """The variant name of the returned flag value."""
    @property
    def value(self) -> google.protobuf.struct_pb2.Struct:
        """The response value of the object flag evaluation, will be unset in the case of error.

        NOTE: This structure will need to be decoded from google/protobuf/struct.proto before it is returned to the SDK
        """

    @property
    def metadata(self) -> google.protobuf.struct_pb2.Struct:
        """Metadata for this evaluation"""

    def __init__(
        self,
        *,
        value: google.protobuf.struct_pb2.Struct | None = ...,
        reason: builtins.str = ...,
        variant: builtins.str = ...,
        metadata: google.protobuf.struct_pb2.Struct | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["metadata", b"metadata", "value", b"value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["metadata", b"metadata", "reason", b"reason", "value", b"value", "variant", b"variant"]) -> None: ...

global___ResolveObjectResponse = ResolveObjectResponse

@typing.final
class EventStreamResponse(google.protobuf.message.Message):
    """Response body for the EventStream stream response"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TYPE_FIELD_NUMBER: builtins.int
    DATA_FIELD_NUMBER: builtins.int
    type: builtins.str
    """String key indicating the type of event that is being received, for example, provider_ready or configuration_change"""
    @property
    def data(self) -> google.protobuf.struct_pb2.Struct:
        """Object structure for use when sending relevant metadata to provide context to the event.
        Can be left unset when it is not required.
        """

    def __init__(
        self,
        *,
        type: builtins.str = ...,
        data: google.protobuf.struct_pb2.Struct | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["data", b"data"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["data", b"data", "type", b"type"]) -> None: ...

global___EventStreamResponse = EventStreamResponse

@typing.final
class EventStreamRequest(google.protobuf.message.Message):
    """Empty stream request body"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___EventStreamRequest = EventStreamRequest



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
*
DEPRECATED; use flagd.sync.v1
Flag definition sync API

This proto defines a simple API to synchronize a feature flag definition.
It supports establishing a stream for getting notifications about changes in a flag definition.
"""

import abc
import collections.abc
import grpc
import grpc.aio
import openfeature.schemas.protobuf.sync.v1.sync_service_pb2
import typing

_T = typing.TypeVar("_T")

class _MaybeAsyncIterator(collections.abc.AsyncIterator[_T], collections.abc.Iterator[_T], metaclass=abc.ABCMeta): ...

class _ServicerContext(grpc.ServicerContext, grpc.aio.ServicerContext):  # type: ignore[misc, type-arg]
    ...

class FlagSyncServiceStub:
    """FlagService implements a server streaming to provide realtime flag configurations"""

    def __init__(self, channel: typing.Union[grpc.Channel, grpc.aio.Channel]) -> None: ...
    SyncFlags: grpc.UnaryStreamMultiCallable[
        openfeature.schemas.protobuf.sync.v1.sync_service_pb2.SyncFlagsRequest,
        openfeature.schemas.protobuf.sync.v1.sync_service_pb2.SyncFlagsResponse,
    ]

    FetchAllFlags: grpc.UnaryUnaryMultiCallable[
        openfeature.schemas.protobuf.sync.v1.sync_service_pb2.FetchAllFlagsRequest,
        openfeature.schemas.protobuf.sync.v1.sync_service_pb2.FetchAllFlagsResponse,
    ]

class FlagSyncServiceAsyncStub:
    """FlagService implements a server streaming to provide realtime flag configurations"""

    SyncFlags: grpc.aio.UnaryStreamMultiCallable[
        openfeature.schemas.protobuf.sync.v1.sync_service_pb2.SyncFlagsRequest,
        openfeature.schemas.protobuf.sync.v1.sync_service_pb2.SyncFlagsResponse,
    ]

    FetchAllFlags: grpc.aio.UnaryUnaryMultiCallable[
        openfeature.schemas.protobuf.sync.v1.sync_service_pb2.FetchAllFlagsRequest,
        openfeature.schemas.protobuf.sync.v1.sync_service_pb2.FetchAllFlagsResponse,
    ]

class FlagSyncServiceServicer(metaclass=abc.ABCMeta):
    """FlagService implements a server streaming to provide realtime flag configurations"""

    @abc.abstractmethod
    def SyncFlags(
        self,
        request: openfeature.schemas.protobuf.sync.v1.sync_service_pb2.SyncFlagsRequest,
        context: _ServicerContext,
    ) -> typing.Union[collections.abc.Iterator[openfeature.schemas.protobuf.sync.v1.sync_service_pb2.SyncFlagsResponse], collections.abc.AsyncIterator[openfeature.schemas.protobuf.sync.v1.sync_service_pb2.SyncFlagsResponse]]: ...

    @abc.abstractmethod
    def FetchAllFlags(
        self,
        request: openfeature.schemas.protobuf.sync.v1.sync_service_pb2.FetchAllFlagsRequest,
        context: _ServicerContext,
    ) -> typing.Union[openfeature.schemas.protobuf.sync.v1.sync_service_pb2.FetchAllFlagsResponse, collections.abc.Awaitable[openfeature.schemas.protobuf.sync.v1.sync_service_pb2.FetchAllFlagsResponse]]: ...

def add_FlagSyncServiceServicer_to_server(servicer: FlagSyncServiceServicer, server: typing.Union[grpc.Server, grpc.aio.Server]) -> None: ...



"""
# This is a Python Provider to interact with flagd
#
# -- Usage --
# open_feature_api.set_provider(flagd_provider.FlagdProvider())
# flag_value =  open_feature_client.get_string_value(
#                   key="foo",
#                   default_value="missingflag"
#               )
# print(f"Flag Value is: {flag_value}")
#   OR the more verbose option
# flag = open_feature_client.get_string_details(key="foo", default_value="missingflag")
# print(f"Flag is: {flag.value}")
#   OR
# print(f"Flag Details: {vars(flag)}"")
#
# -- Customisation --
# Follows flagd defaults: 'http' protocol on 'localhost' on port '8013'
# But can be overridden:
# provider = open_feature_api.get_provider()
# provider.initialise(schema="https",endpoint="example.com",port=1234,timeout=10)
"""

import typing
import warnings

import grpc

from openfeature.evaluation_context import EvaluationContext
from openfeature.event import ProviderEventDetails
from openfeature.flag_evaluation import FlagResolutionDetails
from openfeature.hook import Hook
from openfeature.provider import AbstractProvider
from openfeature.provider.metadata import Metadata

from .config import CacheType, Config, ResolverType
from .resolvers import AbstractResolver, GrpcResolver, InProcessResolver
from .sync_metadata_hook import SyncMetadataHook

T = typing.TypeVar("T")


class FlagdProvider(AbstractProvider):
    """Flagd OpenFeature Provider"""

    def __init__(  # noqa: PLR0913
        self,
        host: typing.Optional[str] = None,
        port: typing.Optional[int] = None,
        tls: typing.Optional[bool] = None,
        deadline_ms: typing.Optional[int] = None,
        timeout: typing.Optional[int] = None,
        retry_backoff_ms: typing.Optional[int] = None,
        selector: typing.Optional[str] = None,
        provider_id: typing.Optional[str] = None,
        resolver_type: typing.Optional[ResolverType] = None,
        offline_flag_source_path: typing.Optional[str] = None,
        stream_deadline_ms: typing.Optional[int] = None,
        keep_alive_time: typing.Optional[int] = None,
        cache: typing.Optional[CacheType] = None,
        max_cache_size: typing.Optional[int] = None,
        retry_backoff_max_ms: typing.Optional[int] = None,
        retry_grace_period: typing.Optional[int] = None,
        cert_path: typing.Optional[str] = None,
        default_authority: typing.Optional[str] = None,
        channel_credentials: typing.Optional[grpc.ChannelCredentials] = None,
    ):
        """
        Create an instance of the FlagdProvider

        :param host: the host to make requests to
        :param port: the port the flagd service is available on
        :param tls: enable/disable secure TLS connectivity
        :param deadline_ms: the maximum to wait before a request times out
        :param timeout: the maximum time to wait before a request times out
        :param retry_backoff_ms: the number of milliseconds to backoff
        :param offline_flag_source_path: the path to the flag source file
        :param stream_deadline_ms: the maximum time to wait before a request times out
        :param keep_alive_time: the number of milliseconds to keep alive
        :param resolver_type: the type of resolver to use
        """
        if deadline_ms is None and timeout is not None:
            deadline_ms = timeout * 1000
            warnings.warn(
                "'timeout' property is deprecated, please use 'deadline' instead, be aware that 'deadline' is in milliseconds",
                DeprecationWarning,
                stacklevel=2,
            )

        self.config = Config(
            host=host,
            port=port,
            tls=tls,
            deadline_ms=deadline_ms,
            retry_backoff_ms=retry_backoff_ms,
            retry_backoff_max_ms=retry_backoff_max_ms,
            retry_grace_period=retry_grace_period,
            selector=selector,
            provider_id=provider_id,
            resolver=resolver_type,
            offline_flag_source_path=offline_flag_source_path,
            stream_deadline_ms=stream_deadline_ms,
            keep_alive_time=keep_alive_time,
            cache=cache,
            max_cache_size=max_cache_size,
            cert_path=cert_path,
            default_authority=default_authority,
            channel_credentials=channel_credentials,
        )
        self.enriched_context: dict = {}

        self.resolver = self.setup_resolver()
        self.hooks: list[Hook] = [SyncMetadataHook(self.get_enriched_context)]

    def get_enriched_context(self) -> EvaluationContext:
        return EvaluationContext(attributes=self.enriched_context)

    def get_provider_hooks(self) -> list[Hook]:
        return self.hooks

    def setup_resolver(self) -> AbstractResolver:
        if self.config.resolver == ResolverType.RPC:
            return GrpcResolver(
                self.config,
                self.emit_provider_ready,
                self.emit_provider_error,
                self.emit_provider_stale,
                self.emit_provider_configuration_changed,
            )
        elif (
            self.config.resolver == ResolverType.IN_PROCESS
            or self.config.resolver == ResolverType.FILE
        ):
            return InProcessResolver(
                self.config,
                self.emit_provider_ready_with_context,
                self.emit_provider_error,
                self.emit_provider_stale,
                self.emit_provider_configuration_changed,
            )
        else:
            raise ValueError(
                f"`resolver_type` parameter invalid: {self.config.resolver}"
            )

    def initialize(self, evaluation_context: EvaluationContext) -> None:
        self.resolver.initialize(evaluation_context)

    def shutdown(self) -> None:
        if self.resolver:
            self.resolver.shutdown()

    def get_metadata(self) -> Metadata:
        """Returns provider metadata"""
        return Metadata(name="FlagdProvider")

    def resolve_boolean_details(
        self,
        key: str,
        default_value: bool,
        evaluation_context: typing.Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[bool]:
        return self.resolver.resolve_boolean_details(
            key, default_value, evaluation_context
        )

    def resolve_string_details(
        self,
        key: str,
        default_value: str,
        evaluation_context: typing.Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[str]:
        return self.resolver.resolve_string_details(
            key, default_value, evaluation_context
        )

    def resolve_float_details(
        self,
        key: str,
        default_value: float,
        evaluation_context: typing.Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[float]:
        return self.resolver.resolve_float_details(
            key, default_value, evaluation_context
        )

    def resolve_integer_details(
        self,
        key: str,
        default_value: int,
        evaluation_context: typing.Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[int]:
        return self.resolver.resolve_integer_details(
            key, default_value, evaluation_context
        )

    def resolve_object_details(
        self,
        key: str,
        default_value: typing.Union[dict, list],
        evaluation_context: typing.Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[typing.Union[dict, list]]:
        return self.resolver.resolve_object_details(
            key, default_value, evaluation_context
        )

    def emit_provider_ready_with_context(
        self, details: ProviderEventDetails, context: dict
    ) -> None:
        self.enriched_context = context
        self.emit_provider_ready(details)
        pass



import typing

from typing_extensions import Protocol

from openfeature.evaluation_context import EvaluationContext
from openfeature.flag_evaluation import FlagResolutionDetails


class AbstractResolver(Protocol):
    def initialize(self, evaluation_context: EvaluationContext) -> None: ...

    def shutdown(self) -> None: ...

    def resolve_boolean_details(
        self,
        key: str,
        default_value: bool,
        evaluation_context: typing.Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[bool]: ...

    def resolve_string_details(
        self,
        key: str,
        default_value: str,
        evaluation_context: typing.Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[str]: ...

    def resolve_float_details(
        self,
        key: str,
        default_value: float,
        evaluation_context: typing.Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[float]: ...

    def resolve_integer_details(
        self,
        key: str,
        default_value: int,
        evaluation_context: typing.Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[int]: ...

    def resolve_object_details(
        self,
        key: str,
        default_value: typing.Union[dict, list],
        evaluation_context: typing.Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[typing.Union[dict, list]]: ...



import time
import typing

from json_logic import builtins, jsonLogic
from json_logic.types import JsonValue

from openfeature.evaluation_context import EvaluationContext

from .custom_ops import (
    ends_with,
    fractional,
    sem_ver,
    starts_with,
)

OPERATORS = {
    **builtins.BUILTINS,
    "fractional": fractional,
    "starts_with": starts_with,
    "ends_with": ends_with,
    "sem_ver": sem_ver,
}


def targeting(
    key: str,
    targeting: dict,
    evaluation_context: typing.Optional[EvaluationContext] = None,
) -> JsonValue:
    json_logic_context = evaluation_context.attributes if evaluation_context else {}
    json_logic_context["$flagd"] = {"flagKey": key, "timestamp": int(time.time())}
    json_logic_context["targetingKey"] = (
        evaluation_context.targeting_key if evaluation_context else None
    )
    return jsonLogic(targeting, json_logic_context, OPERATORS)



import typing

import pytest
from pytest_bdd import given, parsers, when
from tests.e2e.parsers import to_bool, to_list

from openfeature.evaluation_context import EvaluationContext

from ._utils import type_cast


@pytest.fixture
def evaluation_context() -> EvaluationContext:
    return EvaluationContext()


@given(
    parsers.cfparse(
        'a context containing a targeting key with value "{targeting_key}"'
    ),
)
def assign_targeting_context(evaluation_context: EvaluationContext, targeting_key: str):
    """a context containing a targeting key with value <targeting key>."""
    evaluation_context.targeting_key = targeting_key


@given(
    parsers.cfparse(
        'a context containing a key "{key}", with type "{type_info}" and with value "{value}"'
    ),
)
def update_context(
    evaluation_context: EvaluationContext, key: str, type_info: str, value: str
):
    """a context containing a key and value."""
    evaluation_context.attributes[key] = type_cast[type_info](value)


@when(
    parsers.cfparse(
        'context contains keys {fields:s} with values "{svalue}", "{svalue2}", {ivalue:d}, "{bvalue:bool}"',
        extra_types={"bool": to_bool, "s": to_list},
    ),
)
def assign_targeting_context_2(
    evaluation_context: EvaluationContext,
    fields: list,
    svalue: str,
    svalue2: str,
    ivalue: int,
    bvalue: bool,
):
    evaluation_context.attributes[fields[0]] = svalue
    evaluation_context.attributes[fields[1]] = svalue2
    evaluation_context.attributes[fields[2]] = ivalue
    evaluation_context.attributes[fields[3]] = bvalue


@given(
    parsers.cfparse(
        'a context containing a nested property with outer key "{outer}" and inner key "{inner}", with value "{value}"'
    ),
)
def update_context_nested(
    evaluation_context: EvaluationContext,
    outer: str,
    inner: str,
    value: typing.Union[str, int],
):
    """a context containing a nested property with outer key, and inner key, and value."""
    if outer not in evaluation_context.attributes:
        evaluation_context.attributes[outer] = {}
    evaluation_context.attributes[outer][inner] = value



import pytest

from openfeature.contrib.provider.flagd.config import ResolverType
from tests.e2e.testfilter import TestFilter

resolver = ResolverType.RPC
feature_list = ["~targetURI", "~unixsocket", "~sync"]


def pytest_collection_modifyitems(config, items):
    test_filter = TestFilter(
        config, feature_list=feature_list, resolver=resolver.value, base_path=__file__
    )
    test_filter.filter_items(items)


@pytest.fixture()
def resolver_type() -> ResolverType:
    return resolver



import os
from unittest.mock import Mock

import pytest

from openfeature import api
from openfeature.contrib.provider.flagd import FlagdProvider
from openfeature.contrib.provider.flagd.config import Config
from openfeature.contrib.provider.flagd.resolvers.process.connector.file_watcher import (
    FileWatcher,
)
from openfeature.contrib.provider.flagd.resolvers.process.flags import Flag, FlagStore


def create_client(provider: FlagdProvider):
    api.set_provider(provider)
    return api.get_client()


@pytest.mark.parametrize(
    "file_name",
    [
        "basic-flag.json",
        "basic-flag.yaml",
    ],
)
def test_file_load(file_name: str):
    emit_provider_configuration_changed = Mock()
    emit_provider_ready = Mock()
    emit_provider_error = Mock()
    flag_store = FlagStore(emit_provider_configuration_changed)
    path = os.path.abspath(os.path.join(os.path.dirname(__file__), "./flags/"))
    file_watcher = FileWatcher(
        Config(
            offline_flag_source_path=f"{path}/{file_name}",
        ),
        flag_store,
        emit_provider_ready,
        emit_provider_error,
    )
    file_watcher.initialize(None)

    flag = flag_store.get_flag("basic-flag")

    assert flag is not None
    assert isinstance(flag, Flag)
